"""
Context Broker Module for API Client
"""

import re
import json
import os
from math import inf
from typing import Any, Dict, List, Union, Optional, Literal
from urllib.parse import urljoin
import requests
from pydantic import TypeAdapter, PositiveInt, PositiveFloat
from filip.clients.base_http_client import BaseHttpClient, NgsiURLVersion
from filip.config import settings
from filip.models.base import FiwareLDHeader, PaginationMethod, core_context
from filip.models.ngsi_v2.base import AttrsFormat
from filip.models.ngsi_ld.subscriptions import SubscriptionLD
from filip.models.ngsi_ld.context import (
    DataTypeLD,
    ContextLDEntity,
    ContextLDEntityKeyValues,
    ContextProperty,
    ContextRelationship,
    NamedContextProperty,
    NamedContextRelationship,
    ActionTypeLD,
    UpdateLD,
)
from filip.models.ngsi_v2.context import Query


class ContextBrokerLDClient(BaseHttpClient):
    """
    Implementation of NGSI-LD Context Broker functionalities, such as creating
    entities and subscriptions; retrieving, updating and deleting data.
    Further documentation:
    https://fiware-orion.readthedocs.io/en/master/

    Api specifications for LD are located here:
    https://www.etsi.org/deliver/etsi_gs/CIM/001_099/009/01.04.01_60/gs_cim009v010401p.pdf
    """

    def __init__(
        self,
        url: str = None,
        *,
        session: requests.Session = None,
        fiware_header: FiwareLDHeader = None,
        **kwargs,
    ):
        """

        Args:
            url: Url of context broker server
            session (requests.Session):
            fiware_header (FiwareHeader): fiware service and fiware service path
            **kwargs (Optional): Optional arguments that ``request`` takes.
        """
        # set service url
        url = url or settings.LD_CB_URL
        # base_http_client overwrites empty header with FiwareHeader instead of FiwareLD
        init_header = fiware_header if fiware_header else FiwareLDHeader()
        if init_header.link_header is None:
            init_header.set_context(core_context)
        super().__init__(url=url, session=session, fiware_header=init_header, **kwargs)
        # set the version specific url-pattern
        self._url_version = NgsiURLVersion.ld_url.value
        # For uplink requests, the Content-Type header is essential,
        # Accept will be ignored
        # For downlink requests, the Accept header is essential,
        # Content-Type will be ignored

        # default uplink content JSON
        self.headers.update({"Content-Type": "application/json"})
        # default downlink content JSON-LD
        self.headers.update({"Accept": "application/ld+json"})

        if init_header.ngsild_tenant is not None:
            self.__make_tenant()

    def __pagination(
        self,
        *,
        method: PaginationMethod = PaginationMethod.GET,
        url: str,
        headers: Dict,
        limit: Union[PositiveInt, PositiveFloat] = None,
        params: Dict = None,
        data: str = None,
    ) -> List[Dict]:
        """
        NGSIv2 implements a pagination mechanism in order to help clients to
        retrieve large sets of resources. This mechanism works for all listing
        operations in the API (e.g. GET /v2/entities, GET /v2/subscriptions,
        POST /v2/op/query, etc.). This function helps getting datasets that are
        larger than the limit for the different GET operations.

        https://fiware-orion.readthedocs.io/en/master/user/pagination/index.html

        Args:
            url: Information about the url, obtained from the original function
            headers: The headers from the original function
            params:
            limit:

        Returns:
            object:

        """
        if limit is None:
            limit = inf
        if limit > 1000:
            params["limit"] = 1000  # maximum items per request
        else:
            params["limit"] = limit

        if self.session:
            session = self.session
        else:
            session = requests.Session()
        with session:
            res = session.request(
                method=method, url=url, params=params, headers=headers, data=data
            )
            if res.ok:
                items = res.json()
                # do pagination
                if self._url_version == NgsiURLVersion.v2_url.value:
                    count = int(res.headers["Fiware-Total-Count"])
                elif self._url_version == NgsiURLVersion.ld_url.value:
                    count = int(res.headers["NGSILD-Results-Count"])
                else:
                    count = 0

                while len(items) < limit and len(items) < count:
                    # Establishing the offset from where entities are retrieved
                    params["offset"] = len(items)
                    params["limit"] = min(1000, (limit - len(items)))
                    res = session.request(
                        method=method,
                        url=url,
                        params=params,
                        headers=headers,
                        data=data,
                    )
                    if res.ok:
                        items.extend(res.json())
                    else:
                        res.raise_for_status()
                self.logger.debug("Received: %s", items)
                return items
            res.raise_for_status()

    def get_version(self) -> Dict:
        """
        Gets version of Orion-LD context broker
        Returns:
            Dictionary with response
        """
        url = urljoin(self.base_url, "/version")
        try:
            res = self.get(url=url)
            if res.ok:
                return res.json()
            res.raise_for_status()
        except requests.RequestException as err:
            self.logger.error(err)
            raise

    def __make_tenant(self):
        """
        Create tenant if tenant
        is given in headers
        """
        idhex = f"urn:ngsi-ld:{os.urandom(6).hex()}"
        e = ContextLDEntity(id=idhex, type="TemporaryTenant")
        try:
            self.post_entity(entity=e)
            self.delete_entity_by_id(idhex)
        except Exception as err:
            self.log_error(err=err, msg="Error while creating tenant")
            raise

    def get_statistics(self) -> Dict:
        """
        Gets statistics of context broker
        Returns:
            Dictionary with response
        """
        url = urljoin(self.base_url, "statistics")
        try:
            res = self.get(url=url)
            if res.ok:
                return res.json()
            res.raise_for_status()
        except requests.RequestException as err:
            self.logger.error(err)
            raise

    def post_entity(
        self,
        entity: Union[ContextLDEntity, ContextLDEntityKeyValues],
        append: bool = False,
        update: bool = False,
    ):
        """
        Function registers an Object with the NGSI-LD Context Broker,
        if it already exists it can be automatically updated
        if the update flag bool is True.
        First a post request with the entity is tried, if the response code
        is 422 the entity is uncrossable, as it already exists there are two
        options, either overwrite it, if the attribute have changed
        (e.g. at least one new/new values) (update = True) or leave
        it the way it is (update=False)

        """
        url = urljoin(self.base_url, f"{self._url_version}/entities")
        headers = self.headers.copy()
        if isinstance(entity, ContextLDEntityKeyValues):
            entity = entity.to_entity()
        if entity.model_dump().get("@context", None) is not None:
            headers.update({"Content-Type": "application/ld+json"})
            headers.update({"Link": None})
        try:
            res = self.post(
                url=url,
                headers=headers,
                json=entity.model_dump(
                    exclude_unset=False, exclude_defaults=False, exclude_none=True
                ),
            )
            if res.ok:
                self.logger.info("Entity successfully posted!")
                return res.headers.get("Location")
            res.raise_for_status()
        except requests.RequestException as err:
            if err.response is not None and err.response.status_code == 409:
                if append:  # 409 entity already exists
                    return self.append_entity_attributes(entity=entity)
                elif update:
                    return self.override_entities(entities=[entity])
            msg = f"Could not post entity {entity.id}"
            self.log_error(err=err, msg=msg)
            raise

    def override_entities(
        self, entities: List[Union[ContextLDEntity, ContextLDEntityKeyValues]]
    ):
        """
        Function to create or override existing entites with the NGSI-LD Context Broker.
        The batch operation with Upsert will be used.
        """
        return self.entity_batch_operation(
            entities=entities, action_type=ActionTypeLD.UPSERT, options="replace"
        )

    def get_entity(
        self,
        entity_id: str,
        entity_type: str = None,
        attrs: List[str] = None,
        options: Optional[str] = None,
        geometryProperty: Optional[str] = None,
    ) -> Union[ContextLDEntity, ContextLDEntityKeyValues, Dict[str, Any]]:
        """
        This operation must return one entity element only, but there may be
        more than one entity with the same ID (e.g. entities with same ID but
        different types). In such case, an error message is returned, with
        the HTTP status code set to 409 Conflict.

        Args:
            entity_id (String): Id of the entity to be retrieved
            entity_type (String): Entity type, to avoid ambiguity in case
                there are several entities with the same entity id.
            attrs (List of Strings): List of attribute names whose data must be
                included in the response. The attributes are retrieved in the
                order specified by this parameter.
                See "Filtering out attributes and metadata" section for more
                detail. If this parameter is not included, the attributes are
                retrieved in arbitrary order, and all the attributes of the
                entity are included in the response.
                Example: temperature, humidity.
            options (String): keyValues (simplified representation of entity)
                or sysAttrs (include generated attrs createdAt and modifiedAt)
            geometryProperty (String): Name of a GeoProperty. In the case of GeoJSON
                Entity representation, this parameter indicates which GeoProperty to
                use for the "geometry" element. By default, it shall be 'location'.
        Returns:
            ContextEntity
        """
        url = urljoin(self.base_url, f"{self._url_version}/entities/{entity_id}")
        headers = self.headers.copy()
        params = {}
        if entity_type:
            params.update({"type": entity_type})
        if attrs:
            params.update({"attrs": ",".join(attrs)})
        if geometryProperty:
            params.update({"geometryProperty": geometryProperty})
        if options:
            if options != "keyValues" and options != "sysAttrs":
                raise ValueError(
                    f"Only available options are 'keyValues' and 'sysAttrs'"
                )
            params.update({"options": options})

        try:
            res = self.get(url=url, params=params, headers=headers)
            if res.ok:
                self.logger.info("Entity successfully retrieved!")
                self.logger.debug("Received: %s", res.json())
                if options == "keyValues":
                    return ContextLDEntityKeyValues(**res.json())
                else:
                    return ContextLDEntity(**res.json())
            res.raise_for_status()
        except requests.RequestException as err:
            msg = f"Could not load entity {entity_id}"
            self.log_error(err=err, msg=msg)
            raise

    GeometryShape = Literal[
        "Point",
        "MultiPoint",
        "LineString",
        "MultiLineString",
        "Polygon",
        "MultiPolygon",
    ]

    def get_entity_list(
        self,
        entity_id: Optional[str] = None,
        id_pattern: Optional[str] = ".*",
        entity_type: Optional[str] = None,
        attrs: Optional[List[str]] = None,
        q: Optional[str] = None,
        georel: Optional[str] = None,
        geometry: Optional[GeometryShape] = None,
        coordinates: Optional[str] = None,
        geoproperty: Optional[str] = None,
        # csf: Optional[str] = None,  # Context Source Filter
        limit: Optional[PositiveInt] = None,
        options: Optional[str] = None,
    ) -> List[Union[ContextLDEntity, ContextLDEntityKeyValues]]:
        """
        This operation retrieves a list of entities based on different query options.
        By default, the operation retrieves all the entities in the context broker.
        Args:
            entity_id:
                Id of the entity to be retrieved
            id_pattern:
                Regular expression to match the entity id
            entity_type:
                Entity type, to avoid ambiguity in case there are several
                entities with the same entity id.
            attrs:
                List of attribute names whose data must be included in the response.
            q:
                Query expression, composed of attribute names, operators and values.
            georel:
                Geospatial relationship between the query geometry and the entities.
            geometry:
                Type of geometry for the query. The possible values are Point,
                MultiPoint, LineString, MultiLineString, Polygon, MultiPolygon.
            coordinates:
                Coordinates of the query geometry. The coordinates must be
                expressed as a string of comma-separated values.
            geoproperty:
                Name of a GeoProperty. In the case of GeoJSON Entity representation,
                this parameter indicates which GeoProperty to use for the "geometry" element.
            limit:
                Maximum number of entities to retrieve.
            options:
                Further options for the query. The available options are:
                 - keyValues (simplified representation of entity)
                 - sysAttrs (including createdAt and modifiedAt, etc.)
                 - count (include number of all matched entities in response header)
        """
        url = urljoin(self.base_url, f"{self._url_version}/entities/")
        headers = self.headers.copy()
        params = {}
        if entity_id:
            params.update({"id": entity_id})
        if id_pattern:
            params.update({"idPattern": id_pattern})
        if entity_type:
            params.update({"type": entity_type})
        if attrs:
            params.update({"attrs": ",".join(attrs)})
        if q:
            x = re.search(r"[=!<>~]{1}\'.*\'", q.replace(" ", ""))
            if x is not None:
                raise ValueError(
                    f"String/Date/etc. value in {x.group()} must be " f"in double quote"
                )
            params.update({"q": q})
        if georel:
            params.update({"georel": georel})
        if geometry:
            params.update({"geometry": geometry})
        if coordinates:
            params.update({"coordinates": coordinates})
        if geoproperty:
            params.update({"geoproperty": geoproperty})
        # if csf:  # ContextSourceRegistration not supported yet
        #     params.update({'csf': csf})
        if limit:
            if limit > 1000:
                raise ValueError("limit must be an integer value <= 1000")
            params.update({"limit": limit})
        if options:
            if options != "keyValues" and options != "sysAttrs":
                raise ValueError(
                    f"Only available options are 'keyValues' and 'sysAttrs'"
                )
            params.update({"options": options})
        # params.update({'local': 'true'})

        try:
            res = self.get(url=url, params=params, headers=headers)
            if res.ok:
                self.logger.info("Entity successfully retrieved!")
                entity_list: List[Union[ContextLDEntity, ContextLDEntityKeyValues]] = []
                if options == "keyValues":
                    entity_list = [
                        ContextLDEntityKeyValues(**item) for item in res.json()
                    ]
                    return entity_list
                else:
                    entity_list = [ContextLDEntity(**item) for item in res.json()]
                    return entity_list
            res.raise_for_status()
        except requests.RequestException as err:
            msg = f"Could not load entity matching{params}"
            self.log_error(err=err, msg=msg)
            raise

    def replace_existing_attributes_of_entity(
        self,
        entity: Union[ContextLDEntity, ContextLDEntityKeyValues],
        append: bool = False,
    ):
        """
        The attributes previously existing in the entity are removed and
        replaced by the ones in the request.

        Args:
            entity (ContextEntity):
            append (bool):
            options:
        Returns:

        """
        url = urljoin(self.base_url, f"{self._url_version}/entities/{entity.id}/attrs")
        headers = self.headers.copy()
        if isinstance(entity, ContextLDEntityKeyValues):
            entity = entity.to_entity()
        if entity.model_dump().get("@context", None) is not None:
            headers.update({"Content-Type": "application/ld+json"})
            headers.update({"Link": None})
        try:
            res = self.patch(
                url=url,
                headers=headers,
                json=entity.model_dump(exclude={"id", "type"}, exclude_none=True),
            )
            if res.ok:
                self.logger.info(f"Entity {entity.id} successfully " "updated!")
            else:
                res.raise_for_status()
        except requests.RequestException as err:
            if err.response is not None and append and err.response.status_code == 207:
                return self.append_entity_attributes(entity=entity)
            msg = f"Could not replace attribute of entity {entity.id} !"
            self.log_error(err=err, msg=msg)
            raise

    def update_entity_attribute(
        self,
        entity_id: str,
        attr: Union[
            ContextProperty,
            ContextRelationship,
            NamedContextProperty,
            NamedContextRelationship,
        ],
        attr_name: str = None,
    ):
        """
        Updates a specified attribute from an entity.
        Args:
            attr: context attribute to update
            entity_id: Id of the entity. Example: Bcn_Welt
            entity_type: Entity type, to avoid ambiguity in case there are
            several entities with the same entity id.
        """
        headers = self.headers.copy()
        if not isinstance(attr, NamedContextProperty) or not isinstance(
            attr, NamedContextRelationship
        ):
            assert attr_name is not None, (
                "Missing name for attribute. "
                "attr_name must be present if"
                "attr is of type ContextAttribute"
            )
        else:
            assert attr_name is None, (
                "Invalid argument attr_name. Do not set "
                "attr_name if attr is of type "
                "NamedContextAttribute or NamedContextRelationship"
            )

        url = urljoin(
            self.base_url, f"{self._url_version}/entities/{entity_id}/attrs/{attr_name}"
        )

        jsonnn = {}
        if isinstance(attr, list) or isinstance(attr, NamedContextProperty):
            jsonnn = attr.model_dump(exclude={"name"}, exclude_none=True)
        else:
            prop = attr.model_dump()
            for key, value in prop.items():
                if value and value != "Property":
                    jsonnn[key] = value

        try:
            res = self.patch(url=url, headers=headers, json=jsonnn)
            if res.ok:
                self.logger.info(
                    f"Attribute {attr_name} of {entity_id} successfully updated!"
                )
            else:
                res.raise_for_status()
        except requests.RequestException as err:
            msg = f"Could not update attribute '{attr_name}' of entity {entity_id}"
            self.log_error(err=err, msg=msg)
            raise

    def append_entity_attributes(
        self,
        entity: Union[ContextLDEntity, ContextLDEntityKeyValues],
        options: Optional[str] = None,
    ):
        """
        Append new Entity attributes to an existing Entity within an NGSI-LD system
        Args:
            entity (ContextLDEntity):
                Entity to append attributes to.
            options (str):
                Options for the request. The only available value is
                'noOverwrite'. If set, it will raise 400, if all attributes
                exist already.

        """
        url = urljoin(self.base_url, f"{self._url_version}/entities/{entity.id}/attrs")
        headers = self.headers.copy()
        if entity.model_dump().get("@context", None) is not None:
            headers.update({"Content-Type": "application/ld+json"})
            headers.update({"Link": None})
        params = {}

        if options:
            if options != "noOverwrite":
                raise ValueError(f"The only available value is 'noOverwrite'")
            params.update({"options": options})

        try:
            res = self.post(
                url=url,
                headers=headers,
                params=params,
                json=entity.model_dump(exclude={"id", "type"}, exclude_none=True),
            )
            if res.ok:
                self.logger.info(f"Entity {entity.id} successfully updated!")
            else:
                res.raise_for_status()
        except requests.RequestException as err:
            msg = f"Could not update entity {entity.id}!"
            self.log_error(err=err, msg=msg)
            raise

    # def update_existing_attribute_by_name(self, entity: ContextLDEntity
    #                                       ):
    #     pass

    def delete_entity_by_id(self, entity_id: str, entity_type: Optional[str] = None):
        """
        Deletes an entity by its id. For deleting mulitple entities at once,
        entity_batch_operation() is more efficient.
        Args:
            entity_id:
                ID of entity to delete.
            entity_type:
                Type of entity to delete.
        """
        url = urljoin(self.base_url, f"{self._url_version}/entities/{entity_id}")
        headers = self.headers.copy()
        params = {}

        if entity_type:
            params.update({"type": entity_type})

        try:
            res = self.delete(url=url, headers=headers, params=params)
            if res.ok:
                self.logger.info(f"Entity {entity_id} successfully deleted")
            else:
                res.raise_for_status()
        except requests.RequestException as err:
            msg = f"Could not delete entity {entity_id}"
            self.log_error(err=err, msg=msg)
            raise

    def delete_attribute(self, entity_id: str, attribute_id: str):
        """
        Deletes an attribute from an entity.
        Args:
            entity_id:
                ID of the entity.
            attribute_id:
                Name of the attribute to delete.
        Returns:

        """
        url = urljoin(
            self.base_url,
            f"{self._url_version}/entities/{entity_id}/attrs/{attribute_id}",
        )
        headers = self.headers.copy()

        try:
            res = self.delete(url=url, headers=headers)
            if res.ok:
                self.logger.info(
                    f"Attribute {attribute_id} of Entity {entity_id} successfully deleted"
                )
            else:
                res.raise_for_status()
        except requests.RequestException as err:
            msg = f"Could not delete attribute {attribute_id} of entity {entity_id}"
            self.log_error(err=err, msg=msg)
            raise

    # SUBSCRIPTION API ENDPOINTS
    def get_subscription_list(self, limit: PositiveInt = inf) -> List[SubscriptionLD]:
        """
        Returns a list of all the subscriptions present in the system.
        Args:
            limit: Limit the number of subscriptions to be retrieved
        Returns:
            list of subscriptions
        """
        url = urljoin(self.base_url, f"{self._url_version}/subscriptions/")
        headers = self.headers.copy()
        params = {}

        # We always use the 'count' option to check weather pagination is
        # required
        params.update({"options": "count"})
        try:
            items = self.__pagination(
                limit=limit, url=url, params=params, headers=headers
            )
            adapter = TypeAdapter(List[SubscriptionLD])
            return adapter.validate_python(items)
        except requests.RequestException as err:
            msg = "Could not load subscriptions!"
            self.log_error(err=err, msg=msg)
            raise

    def post_subscription(
        self, subscription: SubscriptionLD, update: bool = False
    ) -> str:
        """
        Creates a new subscription. The subscription is represented by a
        Subscription object defined in filip.cb.models.

        If the subscription already exists, the adding is prevented and the id
        of the existing subscription is returned.

        A subscription is deemed as already existing if there exists a
        subscription with the exact same subject and notification fields. All
        optional fields are not considered.

        Args:
            subscription: Subscription
            update: True - If the subscription already exists, update it
                    False- If the subscription already exists, throw warning
        Returns:
            str: Id of the (created) subscription

        """
        existing_subscriptions = self.get_subscription_list()

        sub_hash = subscription.model_dump_json(
            include={"subject", "notification", "type"}
        )
        for ex_sub in existing_subscriptions:
            if sub_hash == ex_sub.model_dump_json(
                include={"subject", "notification", "type"}
            ):
                self.logger.info("Subscription already exists")
                if update:
                    self.logger.info("Updated subscription")
                    subscription.id = ex_sub.id
                    self.update_subscription(subscription)
                else:
                    self.logger.warning(
                        f"Subscription existed already with the id" f" {ex_sub.id}"
                    )
                return ex_sub.id

        url = urljoin(self.base_url, f"{self._url_version}/subscriptions")
        headers = self.headers.copy()
        if subscription.model_dump().get("@context", None) is not None:
            headers.update({"Content-Type": "application/ld+json"})
            headers.update({"Link": None})
        try:
            res = self.post(
                url=url,
                headers=headers,
                data=subscription.model_dump_json(
                    exclude_unset=False, exclude_defaults=False, exclude_none=True
                ),
            )
            if res.ok:
                self.logger.info("Subscription successfully created!")
                return res.headers["Location"].split("/")[-1]
            res.raise_for_status()
        except requests.RequestException as err:
            msg = "Could not send subscription!"
            self.log_error(err=err, msg=msg)
            raise

    def get_subscription(self, subscription_id: str) -> SubscriptionLD:
        """
        Retrieves a subscription from the context broker.
        Args:
            subscription_id: id of the subscription

        Returns:

        """
        url = urljoin(
            self.base_url, f"{self._url_version}/subscriptions/{subscription_id}"
        )
        headers = self.headers.copy()
        try:
            res = self.get(url=url, headers=headers)
            if res.ok:
                self.logger.debug("Received: %s", res.json())
                return SubscriptionLD(**res.json())
            res.raise_for_status()
        except requests.RequestException as err:
            msg = f"Could not load subscription {subscription_id}"
            self.log_error(err=err, msg=msg)
            raise

    def update_subscription(self, subscription: SubscriptionLD) -> None:
        """
        Only the fields included in the request are updated in the subscription.
        Args:
            subscription: Subscription to update
        Returns:

        """
        url = urljoin(
            self.base_url, f"{self._url_version}/subscriptions/{subscription.id}"
        )
        headers = self.headers.copy()
        if subscription.model_dump().get("@context", None) is not None:
            headers.update({"Content-Type": "application/ld+json"})
            headers.update({"Link": None})
        try:
            res = self.patch(
                url=url,
                headers=headers,
                data=subscription.model_dump_json(
                    exclude={"id"},
                    exclude_none=True,
                ),
            )
            if res.ok:
                self.logger.info("Subscription successfully updated!")
            else:
                res.raise_for_status()
        except requests.RequestException as err:
            msg = f"Could not update subscription {subscription.id}"
            self.log_error(err=err, msg=msg)
            raise

    def delete_subscription(self, subscription_id: str) -> None:
        """
        Deletes a subscription from a Context Broker
        Args:
            subscription_id: id of the subscription
        """
        url = urljoin(
            self.base_url, f"{self._url_version}/subscriptions/{subscription_id}"
        )
        headers = self.headers.copy()
        try:
            res = self.delete(url=url, headers=headers)
            if res.ok:
                self.logger.info(
                    f"Subscription '{subscription_id}' " f"successfully deleted!"
                )
            else:
                res.raise_for_status()
        except requests.RequestException as err:
            msg = f"Could not delete subscription {subscription_id}"
            self.log_error(err=err, msg=msg)
            raise

    def log_multi_errors(self, errors: List[Dict]) -> None:
        for error in errors:
            entity_id = error["entityId"]
            error_details: dict = error["error"]
            error_title = error_details.get("title")
            error_status = error_details.get("status")
            # error_detail = error_details['detail']
            self.logger.error(
                "Response status: %d, Entity: %s, Reason: %s",
                error_status,
                entity_id,
                error_title,
            )

    def handle_multi_status_response(self, res: requests.Response):
        """
        Handles the response of a batch_operation. If the response contains
        errors, they are logged. If the response contains only errors, a RuntimeError
        is raised.
        Args:
            res:

        Returns:

        """
        try:
            res.raise_for_status()
            if res.text:
                response_data = res.json()
                if "errors" in response_data:
                    errors = response_data["errors"]
                    self.log_multi_errors(errors)
                if "success" in response_data:
                    successList = response_data["success"]
                    if len(successList) == 0:
                        raise RuntimeError(
                            "Batch operation resulted in errors only, see logs"
                        )
            else:
                self.logger.info("Empty response received.")
        except json.JSONDecodeError:
            self.logger.info(
                "Error decoding JSON. Response may not be in valid JSON format."
            )

    # Batch operation API
    def entity_batch_operation(
        self,
        *,
        entities: List[Union[ContextLDEntity, ContextLDEntityKeyValues]],
        action_type: Union[ActionTypeLD, str],
        options: Literal["noOverwrite", "replace", "update"] = None,
    ) -> None:
        """
        This operation allows to create, update and/or delete several entities
        in a single batch operation.

        This operation is split in as many individual operations as entities
        in the entities vector, so the actionType is executed for each one of
        them. Depending on the actionType, a mapping with regular non-batch
        operations can be done:

        append: maps to POST /v2/entities (if the entity does not already exist)
        or POST /v2/entities/<id>/attrs (if the entity already exists).

        appendStrict: maps to POST /v2/entities (if the entity does not
        already exist) or POST /v2/entities/<id>/attrs?options=append (if the
        entity already exists).

        update: maps to PATCH /v2/entities/<id>/attrs.

        delete: maps to DELETE /v2/entities/<id>/attrs/<attrName> on every
            attribute included in the entity or to DELETE /v2/entities/<id> if
            no attribute were included in the entity.

        replace: maps to PUT /v2/entities/<id>/attrs.

        Args:
            entities: "an array of entities, each entity specified using the "
                      "JSON entity representation format "
            action_type (Update): "actionType, to specify the kind of update
                    action to do: either append, appendStrict, update, delete,
                    or replace. "
            options (str): Optional 'noOverwrite' 'replace' 'update'

        Returns:

        """

        url = urljoin(
            self.base_url, f"{self._url_version}/entityOperations/{action_type.value}"
        )

        headers = self.headers.copy()
        ctx = any(e.model_dump().get("@context", None) is not None for e in entities)

        nctx = any(e.model_dump().get("@context", None) is None for e in entities)
        if ctx and not nctx:
            headers.update({"Content-Type": "application/ld+json"})
            headers.update({"Link": None})
        elif not ctx and nctx:
            headers.update({"Content-Type": "application/json"})
        else:
            self.logger.warning(
                "Detected mixed context provision in batch operation: "
                "Some entities have @context field while others don't. "
                "FiLiP use application/json and Link header by default, so that "
                "the entities with @context will be rejected by CB"
            )

        params = {}
        if options:
            params.update({"options": options})
        update = UpdateLD(entities=entities)
        try:
            if action_type == ActionTypeLD.DELETE:
                id_list = [entity.id for entity in entities]
                res = self.post(
                    url=url, headers=headers, params=params, data=json.dumps(id_list)
                )
            else:
                res = self.post(
                    url=url,
                    headers=headers,
                    params=params,
                    data=json.dumps(
                        update.model_dump(
                            by_alias=True,
                            exclude_none=True,
                        ).get("entities")
                    ),
                )
            self.handle_multi_status_response(res)
        except RuntimeError as rerr:
            raise rerr
        except Exception as err:
            raise err
        else:
            self.logger.info(f"Update operation {action_type} succeeded!")

    def validate_relationship(
        self,
        relationship: Union[
            NamedContextProperty,
            ContextProperty,
            NamedContextRelationship,
            ContextRelationship,
            Dict,
        ],
    ) -> bool:
        """
        Validates a relationship. A relationship is valid if it points to an existing
        entity. Otherwise, it is considered invalid

        Args:
            relationship: relationship to validate,assumed to be property or relationship
            since there is no geoproperty with string value
        Returns
            True if the relationship is valid, False otherwise
        """
        if hasattr(relationship, "value"):
            destination_id = relationship.value
        elif hasattr(relationship, "object"):
            destination_id = relationship.object
        elif isinstance(relationship, dict):
            _sentinel = object()
            destination_id = relationship.get("value", _sentinel)
            if destination_id is _sentinel:
                raise ValueError(
                    "Invalid ld relationship dictionary format\n"
                    "Expected format: {"
                    f'"type": "{DataTypeLD.RELATIONSHIP[0]}", '
                    '"value" "entity_id"}'
                )
        else:
            raise ValueError("Invalid relationship type.")
        try:
            destination_entity = self.get_entity(entity_id=destination_id)
            return destination_entity.id == destination_id
        except requests.RequestException as err:
            if err.response.status_code == 404:
                return False



"""
Shared models that are used by multiple submodules
"""

import json

from aenum import Enum
from geojson_pydantic import (
    Point,
    MultiPoint,
    LineString,
    MultiLineString,
    Polygon,
    MultiPolygon,
    Feature,
    FeatureCollection,
)
from pydantic import (
    field_validator,
    model_validator,
    ConfigDict,
    AnyHttpUrl,
    BaseModel,
    Field,
    model_serializer,
    SerializationInfo,
    ValidationInfo,
)

from typing import Union, Optional, Pattern, List, Dict, Any

from filip.models.base import DataType
from filip.models.ngsi_v2.units import validate_unit_data, Unit
from filip.utils.simple_ql import QueryString, QueryStatement
from filip.utils.validators import (
    validate_escape_character_free,
    validate_fiware_datatype_string_protect,
    validate_fiware_datatype_standard,
    validate_fiware_attribute_value_regex,
    validate_fiware_attribute_name_regex,
)


class Http(BaseModel):
    """
    Model for notification and registrations sent or retrieved via HTTP
    """

    url: AnyHttpUrl = Field(
        description="URL referencing the service to be invoked when a "
        "notification is generated. An NGSIv2 compliant server "
        "must support the http URL schema. Other schemas could "
        "also be supported."
    )


class EntityPattern(BaseModel):
    """
    Entity pattern used to create subscriptions or registrations
    """

    id: Optional[str] = Field(default=None, pattern=r"\w")
    idPattern: Optional[Pattern] = None
    type: Optional[str] = Field(default=None, pattern=r"\w")
    typePattern: Optional[Pattern] = None

    @model_validator(mode="after")
    def validate_conditions(self):
        assert (self.id and not self.idPattern) or (not self.id and self.idPattern), (
            "Both cannot be used at the same time, but one of 'id' or "
            "'idPattern must' be present."
        )
        if self.type or self.model_dump().get("typePattern", None):
            assert (self.type and not self.typePattern) or (
                not self.type and self.typePattern
            ), (
                "Type or pattern of the affected entities. "
                "Both cannot be used at the same time."
            )
        return self


class Status(str, Enum):
    """
    Current status of a subscription or registrations
    """

    _init_ = "value __doc__"

    ACTIVE = "active", "for active subscriptions"
    INACTIVE = "inactive", "for inactive subscriptions"
    FAILED = "failed", "for failed subscription"
    EXPIRED = "expired", "for expired subscription"


class Expression(BaseModel):
    """
    By means of a filtering expression, allows to express what is the scope
    of the data provided.
    https://telefonicaid.github.io/fiware-orion/api/v2/stable
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    q: Optional[Union[str, QueryString]] = Field(
        default=None,
        title="Simple Query Language: filter",
        description="If filtering by attribute value (i.e. the expression is "
        "used in a q query), the rest of tokens (if present) "
        "represent the path to a sub-property of the target NGSI "
        "attribute value (which should be a JSON object). Such "
        "sub-property is defined as the target property.",
    )
    mq: Optional[Union[str, QueryString]] = Field(
        default=None,
        title="Simple Query Language: metadata filters",
        description="If filtering by metadata (i.e. the expression is used in "
        "a mq query), the second token represents a metadata name "
        "associated to the target NGSI attribute, target "
        "metadata, and the rest of tokens (if present) represent "
        "the path to a sub-property of the target metadata value "
        "(which should be a JSON object). Such sub-property is "
        "defined as the target property. ",
    )
    georel: Optional[Union[str, QueryString]] = Field(
        default=None,
        title="Metadata filters",
        description="Any of the geographical relationships as specified by "
        "the Geoqueries section of this specification.",
    )
    geometry: Optional[Union[str, QueryString]] = Field(
        default=None,
        title="Metadata filters",
        description="Any of the supported geometries as specified by the "
        "Geoqueries section of this specification.",
    )
    coords: Optional[Union[str, QueryString]] = Field(
        default=None,
        title="Metadata filters",
        description="String representation of coordinates as specified by the "
        "Geoqueries section of the specification.",
    )

    @field_validator("q", "mq")
    @classmethod
    def validate_expressions(cls, v):
        if isinstance(v, str):
            return QueryString.parse_str(v)

    @model_serializer(mode="wrap")
    def serialize(self, serializer: Any, info: SerializationInfo):
        if isinstance(self.q, (QueryString, QueryStatement)):
            self.q = self.q.to_str()
        if isinstance(self.mq, (QueryString, QueryStatement)):
            self.mq = self.mq.to_str()
        if isinstance(self.coords, (QueryString, QueryStatement)):
            self.coords = self.coords.to_str()
        if isinstance(self.georel, (QueryString, QueryStatement)):
            self.georel = self.georel.to_str()
        if isinstance(self.geometry, (QueryString, QueryStatement)):
            self.geometry = self.geometry.to_str()
        return serializer(self)


class AttrsFormat(str, Enum):
    """
    Allowed options for attribute formats
    """

    _init_ = "value __doc__"

    NORMALIZED = "normalized", "Normalized message representation"
    KEY_VALUES = (
        "keyValues",
        "Key value message representation."
        "This mode represents the entity "
        "attributes by their values only, leaving out "
        "the information about type and metadata. "
        "See example below."
        "Example: "
        "{"
        "  'id': 'R12345',"
        "  'type': 'Room',"
        "  'temperature': 22"
        "}",
    )
    VALUES = (
        "values",
        "Key value message representation. "
        "This mode represents the entity as an array of "
        "attribute values. Information about id and type is "
        "left out. See example below. The order of the "
        "attributes in the array is specified by the attrs "
        "URI param (e.g. attrs=branch,colour,engine). "
        "If attrs is not used, the order is arbitrary. "
        "Example:"
        "[ 'Ford', 'black', 78.3 ]",
    )


# NGSIv2 entity models
class Metadata(BaseModel):
    """
    Context metadata is used in FIWARE NGSI in several places, one of them being
    an optional part of the attribute value as described above. Similar to
    attributes, each piece of metadata has.

    Note:
         In NGSI it is not foreseen that metadata may contain nested metadata.
    """

    type: Optional[Union[DataType, str]] = Field(
        default=None,
        title="metadata type",
        description="a metadata type, describing the NGSI value type of the "
        "metadata value Allowed characters "
        "are the ones in the plain ASCII set, except the following "
        "ones: control characters, whitespace, &, ?, / and #.",
        max_length=256,
        min_length=1,
    )
    valid_type = field_validator("type")(validate_fiware_datatype_standard)
    value: Optional[Any] = Field(
        default=None,
        title="metadata value",
        description="a metadata value containing the actual metadata",
    )

    @field_validator("value")
    def validate_value(cls, value, info: ValidationInfo):
        assert json.dumps(value), "metadata not serializable"

        if info.data.get("type").casefold() == "unit":
            value = Unit.model_validate(value)
        return value


class NamedMetadata(Metadata):
    """
    Model for metadata including a name
    """

    name: str = Field(
        title="metadata name",
        description="a metadata name, describing the role of the metadata in "
        "the place where it occurs; for example, the metadata name "
        "accuracy indicates that the metadata value describes how "
        "accurate a given attribute value is. Allowed characters "
        "are the ones in the plain ASCII set, except the following "
        "ones: control characters, whitespace, &, ?, / and #.",
        max_length=256,
        min_length=1,
    )
    valid_name = field_validator("name")(validate_fiware_datatype_standard)

    @model_validator(mode="after")
    def validate_data(self):
        if self.model_dump().get("name", "").casefold() in [
            "unit",
            "unittext",
            "unitcode",
        ]:
            valide_dict = self.model_dump()
            valide_dict.update(validate_unit_data(self.model_dump()))
            return self
        return self

    def to_context_metadata(self):
        return {self.name: Metadata(**self.model_dump())}


class BaseAttribute(BaseModel):
    """
    Model for an attribute is represented by a JSON object with the following
    syntax:

    The attribute NGSI type is specified by the type property, whose value
    is a string containing the NGSI type.

    The attribute metadata is specified by the metadata property. Its value
    is another JSON object which contains a property per metadata element
    defined (the name of the property is the name of the metadata element).
    Each metadata element, in turn, is represented by a JSON object
    containing the following properties:

    Values of entity attributes. For adding it you need to nest it into a
    dict in order to give it a name.

    Example:

        >>> data = {"type": <...>,
                    "metadata": <...>}
        >>> attr = BaseAttribute(**data)

    """

    model_config = ConfigDict(validate_assignment=True)
    type: Union[DataType, str] = Field(
        default=DataType.TEXT,
        description="The attribute type represents the NGSI value type of the "
        "attribute value. Note that FIWARE NGSI has its own type "
        "system for attribute values, so NGSI value types are not "
        "the same as JSON types. Allowed characters "
        "are the ones in the plain ASCII set, except the following "
        "ones: control characters, whitespace, &, ?, / and #.",
        max_length=256,
        min_length=1,
    )
    valid_type = field_validator("type")(validate_fiware_datatype_string_protect)
    metadata: Optional[
        Union[
            Dict[str, Metadata],
            NamedMetadata,
            List[NamedMetadata],
            Dict[str, Dict[str, str]],
        ]
    ] = Field(
        default={},
        title="Metadata",
        description="optional metadata describing properties of the attribute "
        "value like e.g. accuracy, provider, or a timestamp",
    )

    @field_validator("metadata")
    @classmethod
    def validate_metadata_type(cls, value):
        """validator for field 'metadata'"""
        if type(value) == NamedMetadata:
            value = [value]
        elif isinstance(value, dict):
            if all(isinstance(item, Metadata) for item in value.values()):
                value = [
                    NamedMetadata(name=key, **item.model_dump())
                    for key, item in value.items()
                ]
            else:
                json.dumps(value)
                value = [NamedMetadata(name=key, **item) for key, item in value.items()]

        if isinstance(value, list):
            if all(isinstance(item, dict) for item in value):
                value = [NamedMetadata(**item) for item in value]
            if all(isinstance(item, NamedMetadata) for item in value):
                return {
                    item.name: Metadata(**item.model_dump(exclude={"name"}))
                    for item in value
                }

        raise TypeError(f"Invalid type {type(value)}")


class BaseNameAttribute(BaseModel):
    """
    Model to add the name property to an BaseAttribute Model.
    The attribute name describes what kind of property the
    attribute value represents of the entity
    """

    name: str = Field(
        title="Attribute name",
        description="The attribute name describes what kind of property the "
        "attribute value represents of the entity, for example "
        "current_speed. Allowed characters "
        "are the ones in the plain ASCII set, except the following "
        "ones: control characters, whitespace, &, ?, / and #.",
        max_length=256,
        min_length=1,
        # Make it FIWARE-Safe
    )
    valid_name = field_validator("name")(validate_fiware_attribute_name_regex)


class BaseValueAttribute(BaseModel):
    """
    Model to add the value property to an BaseAttribute Model. The Model
    is represented by a JSON object with the following syntax:


    The attribute value is specified by the value property, whose value may
    be any JSON datatype.
    """

    type: Union[DataType, str] = Field(
        default=DataType.TEXT,
        description="The attribute type represents the NGSI value type of the "
        "attribute value. Note that FIWARE NGSI has its own type "
        "system for attribute values, so NGSI value types are not "
        "the same as JSON types. Allowed characters "
        "are the ones in the plain ASCII set, except the following "
        "ones: control characters, whitespace, &, ?, / and #.",
        max_length=256,
        min_length=1,
    )
    valid_type = field_validator("type")(validate_fiware_datatype_string_protect)
    value: Optional[Any] = Field(
        default=None, title="Attribute value", description="the actual data"
    )

    @model_validator(mode="before")
    def validate_value_based_on_type(cls, values):
        type_ = values.get("type")
        value = values.get("value")

        if type_ == DataType.TEXT:
            values["value"] = validate_fiware_attribute_value_regex(str(value))

        return values

    @field_validator("value")
    def validate_value_type(cls, value, info: ValidationInfo):
        """
        Validator for field 'value'
        The validator will try autocast the value based on the given type.
        If `DataType.STRUCTUREDVALUE` is used for type it will also serialize
        pydantic models. With latter operation all additional features of the
        original pydantic model will be dumped.
        If the type is unknown it will check json-serializable.
        """
        type_ = info.data.get("type")
        value_ = value
        if isinstance(value, BaseModel):
            value_ = value.model_dump()
        validate_escape_character_free(value_)

        if value not in (None, "", " "):
            if type_ == DataType.TEXT:
                if isinstance(value, list):
                    return [str(item) for item in value]
                return str(value)
            if type_ == DataType.BOOLEAN:
                if isinstance(value, list):
                    return [bool(item) for item in value]
                return bool(value)
            if type_ in (DataType.NUMBER, DataType.FLOAT):
                if isinstance(value, list):
                    return [float(item) for item in value]
                return float(value)
            if type_ == DataType.INTEGER:
                if isinstance(value, list):
                    return [int(item) for item in value]
                return int(value)
            if type_ == DataType.DATETIME:
                return value
            # allows list
            if type_ == DataType.ARRAY:
                if isinstance(value, list):
                    return value
                raise TypeError(f"{type(value)} does not match " f"{DataType.ARRAY}")
            # allows dict and BaseModel as object
            if type_ == DataType.OBJECT:
                if isinstance(value, dict):
                    value = json.dumps(value)
                    return json.loads(value)
                elif isinstance(value, BaseModel):
                    value.model_dump_json()
                    return value
                raise TypeError(f"{type(value)} does not match " f"{DataType.OBJECT}")

            # allows geojson as structured value
            if type_ == DataType.GEOJSON:
                if isinstance(
                    value,
                    (
                        Point,
                        MultiPoint,
                        LineString,
                        MultiLineString,
                        Polygon,
                        MultiPolygon,
                        Feature,
                        FeatureCollection,
                    ),
                ):
                    return value
                if isinstance(value, dict):
                    _geo_json_type = value.get("type", None)
                    if _geo_json_type == "Point":
                        return Point(**value)
                    elif _geo_json_type == "MultiPoint":
                        return MultiPoint(**value)
                    elif _geo_json_type == "LineString":
                        return LineString(**value)
                    elif _geo_json_type == "MultiLineString":
                        return MultiLineString(**value)
                    elif _geo_json_type == "Polygon":
                        return Polygon(**value)
                    elif _geo_json_type == "MultiPolygon":
                        return MultiPolygon(**value)
                    elif _geo_json_type == "Feature":
                        return Feature(**value)
                    elif _geo_json_type == "FeatureCollection":
                        return FeatureCollection(**value)
                raise TypeError(f"{type(value)} does not match " f"{DataType.GEOJSON}")

            # allows list, dict and BaseModel as structured value
            if type_ == DataType.STRUCTUREDVALUE:
                if isinstance(value, (dict, list)):
                    value = json.dumps(value)
                    return json.loads(value)
                elif isinstance(value, BaseModel):
                    value.model_dump_json()
                    return value
                raise TypeError(
                    f"{type(value)} does not match " f"{DataType.STRUCTUREDVALUE}"
                )

            # if none of the above, check if it is a pydantic model
            if isinstance(value, BaseModel):
                value.model_dump_json()
                return value

            # if none of the above, check if serializable. Hence, no further
            # type check is performed
            value = json.dumps(value)
            return json.loads(value)

        return value



import os
import importlib
import logging
from pathlib import Path
import pickle
from typing import Dict
import pandas as pd
from pandas import DataFrame
from pandas_datapackage_reader import read_datapackage
from filip.utils.validators import validate_http_url

logger = logging.getLogger(__name__)


def load_datapackage(url: str, package_name: str) -> Dict[str, pd.DataFrame]:
    """
    Downloads data package from online source and stores it as hdf-file in
    filip.data named by the <filename>.hdf.

    Args:
        url (str): Valid url to where the data package is hosted
        package_name (str): name of the cached file.

    Returns:
        Dict of dataframes
    """
    # validate arguments
    validate_http_url(url=url)

    # create directory for data if not exists
    validate_http_url(url=url)
    path = Path(__file__).parent.parent.absolute().joinpath("data")
    path.mkdir(parents=True, exist_ok=True)
    package_path = path.joinpath(package_name)

    if os.path.isdir(package_path):
        # read data from filip.data if exists
        logger.info("Found existing data package in 'filip.data'")

        data = {}
        for file in os.listdir(package_path):
            file_name = file[:-4]
            # read in each file as one dataframe, prevents the deletion of NaN
            # values with na_filter=False
            frame = pd.read_csv(
                package_path.joinpath(file), index_col=0, header=0, na_filter=False
            )
            data[file_name] = frame

    else:
        # download external data and store data
        logger.info(
            "Could not find data package in 'filip.data'. Will "
            "try to download from %s",
            url,
        )
        try:
            data = read_datapackage(url)
            # rename keys
            data = {k.replace("-", "_"): v for k, v in data.items()}
            os.mkdir(package_path)

            # store data in filip.data
            for k, v in data.items():
                v: DataFrame = v
                v.loc[:, :] = v[:].applymap(str)
                table_filepath = str(package_path) + f"\\{k.replace('-', '_')}.csv"
                v.to_csv(table_filepath)

        except:
            logger.error("Failed to load data package!")
            raise
    return data



"""
Code generator for data models from schema.json descriptions
"""

import json
import shutil
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Union, Dict, Any, Type
from urllib import parse
from uuid import uuid4
from datamodel_code_generator import InputFileType, generate, ParseResult
from pydantic import create_model

from filip.models.ngsi_v2.context import ContextAttribute, ContextEntity


def create_data_model_file(
    *,
    path: Union[Path, str],
    url: str = None,
    schema: Union[Path, str, ParseResult] = None,
    schema_type: Union[str, InputFileType] = InputFileType.JsonSchema,
    class_name: str = None,
) -> None:
    """
    This will create a data model from data model definitions. The schemas
    can either downloaded from a url or passed as str or dict. Allowed input
    types are defined but the underlying toolbox.

    Many data models suited for FIWARE are located here:
    https://github.com/smart-data-models/data-models

    Args:
        path:
            path where the generated code should saved
        url:
            url to download the definition from
        schema_type (str):
            `auto`, `openapi`, `jsonschema`, `json`, `yaml`, `dict`, `csv`
        class_name:
            classname for the model class

    Returns:
        None

    Examples::

        {
            "type": "object",
            "properties": {
                "number": {
                    "type": "number"
                    },
                "street_name": {
                    "type": "string"
                   },
                "street_type": {
                    "type": "string",
                    "enum": ["Street", "Avenue", "Boulevard"]
                }
            }
        }
    """
    if isinstance(path, str):
        path = Path(path)
    path.parent.mkdir(parents=True, exist_ok=True)

    if isinstance(schema_type, str):
        schema_type = InputFileType(schema_type)

    with TemporaryDirectory() as temp:
        temp = Path(temp)
        output = Path(temp).joinpath(f"{uuid4()}.py")

        if url:
            schema = parse.urlparse(url)
        if not schema:
            raise ValueError(
                "Missing argument! Either 'url' or 'schema' " "must be provided"
            )

        generate(
            input_=schema,
            input_file_type=schema_type,
            output=output,
            class_name=class_name,
        )

        # move temporary file to output directory
        shutil.move(str(output), str(path))


def create_context_entity_model(
    name: str = None,
    data: Dict = None,
    validators: Dict[str, Any] = None,
    path: Union[Path, str] = None,
) -> Type["ContextEntity"]:
    r"""
    Creates a ContextEntity-Model from a dict:

    Args:
        name:
            name of the model
        data:
            dictionary containing the data structure
        validators (optional):
            validators for the new model
        path:
            if present the model will written to *.py file if file ending
            *.py is used and to json-schema if *.json is used.

    Example:

        >>> def username_alphanumeric(cls, value):
                assert v.value.isalnum(), 'must be numeric'
                return value

        >>> model = create_context_entity_model(
                        name='MyModel',
                        data={
                            'id': 'MyId',
                            'type':'MyType',
                            'temp': 'MyProperty'}
                        {'validate_test': validator('temperature')(
                            username_alphanumeric)})

    Returns:
        ContextEntity

    """
    properties = {
        key: (ContextAttribute, ...)
        for key in data.keys()
        if key not in ContextEntity.model_fields
    }
    model = create_model(
        name or "GeneratedContextEntity",
        __base__=ContextEntity,
        __validators__=validators or {},
        **properties,
    )

    # if path exits a file will be generated that contains the model
    if path:
        if isinstance(path, str):
            path = Path(path)

        with TemporaryDirectory() as temp:
            temp = Path(temp)
            output = Path(temp).joinpath(f"{uuid4()}.json")
            output.touch(exist_ok=True)
            with output.open("w") as f:
                json.dump(model.model_json_schema(), f, indent=2)
            if path.suffix == ".json":
                # move temporary file to output directory
                shutil.move(str(output), str(path))
            elif path.suffix == ".py":
                create_data_model_file(path=path, schema=output, class_name=name)
    return model



"""
Implementation of an extended MQTT client that automatically handles the
topic subscription for FIWARE's IoT communication pattern.
"""

import itertools
import logging
import warnings
from datetime import datetime
from typing import Any, Callable, Dict, List, Tuple, Union

import paho.mqtt.client as mqtt

from filip.clients.mqtt.encoder import BaseEncoder, Json, Ultralight
from filip.models.mqtt import IoTAMQTTMessageType
from filip.models.ngsi_v2.iot import (
    Device,
    PayloadProtocol,
    ServiceGroup,
    TransportProtocol,
)


class IoTAMQTTClient(mqtt.Client):
    """
    This class is an extension to the MQTT client from the well established
    Eclipse Paho™ MQTT Python Client. The official documentation is located
    here: https://github.com/eclipse/paho.mqtt.python

    The class adds additional functions to facilitate the communication to
    FIWARE's IoT-Agent via MQTT. It magically generates and subscribes to all
    important topics that are necessary to establish a
    bi-directional communication with the IoT-Agent.

    Note:
        The client does not sync the device configuration with the IoT-Agent.
        This is up to the user!

    Note:
        The extension does not effect the normal workflow or any other
        functionality known from the original client.

        The client does not yet support the retrieval of command
        configurations via mqtt documented here:
        https://fiware-iotagent-json.readthedocs.io/en/latest/usermanual/index.html#api-overview

    Example:
        This example shows the basic usage of the client. It does not
        demonstrate its whole capabilities. Please check the single methods
        for more details. Please also keep in mind that this still requires
        provisioning of the device in the IoT-Agent and sending the commands
        via the context broker. For more details check the additional example
        section::

            from filip.models.ngsi_v2.iot import Device, DeviceAttribute, DeviceCommand, ServiceGroup
            from filip.clients.mqtt import MQTTClient
            from filip.clients.mqtt.encoder import IoTA_Json

            # create a device configuration
            device_attr = DeviceAttribute(name='temperature',
                                          object_id='t',
                                          type="Number")
            device_command = DeviceCommand(name='heater', type="Boolean")
            device = Device(device_id='MyDevice',
                            entity_name='MyDevice',
                            entity_type='Thing',
                            protocol='IoTA-JSON',
                            transport='MQTT',
                            apikey=YourApiKey,
                            attributes=[device_attr],
                            commands=[device_command])

            service_group = ServiceGroup(apikey="YourApiKey", resource="/iot")

            mqttc = MQTTClient(client_id="YourID",
                               userdata=None,
                               protocol=mqtt.MQTTv5,
                               transport="tcp",
                               _devices = [device],
                               service_groups = [service_group])

            # create a callback function that will be called for incoming
            # commands and add it for a single device
            def on_command(client, obj, msg):
                apikey, device_id, payload = \
                    client.get_encoder().decode_message(msg=msg)

                # do_something with the message.
                # For instance write into a queue.

                # acknowledge a command
                client.publish(device_id=device_id,
                               command_name=next(iter(payload))
                               payload=payload)

            mqttc.add_command_callback(on_command)

            # create a non blocking loop
            mqttc.loop_start()

            # publish a multi-measurement for a device
            mqttc.publish(device_id='MyDevice', payload={'t': 50})

            # publish a single measurement for a device
            mqttc.publish(device_id='MyDevice',
                          attribute_name='temperature',
                          payload=50)

            # adding timestamps to measurements using the client


            # adding timestamps to measurements in payload
            from datetime import datetime

            mqttc.publish(device_id='MyDevice',
                          payload={'t': 50,
                                   'timeInstant': datetime.now().astimezone().isoformat()},
                          timestamp=true)

            # stop network loop and disconnect cleanly
            mqttc.loop_stop()
            mqttc.disconnect()

    """

    def __init__(
        self,
        client_id="",
        clean_session=None,
        userdata=None,
        protocol=mqtt.MQTTv311,
        transport="tcp",
        callback_api_version=mqtt.CallbackAPIVersion.VERSION2,
        devices: List[Device] = None,
        service_groups: List[ServiceGroup] = None,
        custom_encoder: Dict[str, BaseEncoder] = None,
    ):
        """
        Args:
            client_id:
                Unique client id string used when connecting
                to the broker. If client_id is zero length or None, then the
                behaviour is defined by which protocol version is in use. If
                using MQTT v3.1.1, then a zero length client id will be sent
                to the broker and the broker will generate a random for the
                client. If using MQTT v3.1 then an id will be randomly
                generated. In both cases, clean_session must be True.
                If this is not the case a ValueError will be raised.
            clean_session:
                boolean that determines the client type. If True,
                the broker will remove all information about this client when it
                disconnects. If False, the client is a persistent client and
                subscription information and queued messages will be retained
                when the client disconnects.
                Note that a client will never discard its own outgoing
                messages on disconnect. Calling connect() or reconnect() will
                cause the messages to be resent.  Use reinitialise() to reset
                a client to its original state. The clean_session argument
                only applies to MQTT versions v3.1.1 and v3.1. It is not
                accepted if the MQTT version is v5.0 - use the clean_start
                argument on connect() instead.
            userdata:
                defined data of any type that is passed as the "userdata"
                parameter to callbacks. It may be updated at a later point
                with the user_data_set() function.
            protocol:
                explicit setting of the MQTT version to use for this client.
                Can be paho.mqtt.client.MQTTv311 (v3.1.1),
                paho.mqtt.client.MQTTv31 (v3.1) or paho.mqtt.client.MQTTv5
                (v5.0), with the default being v3.1.1.
            transport:
                Set to "websockets" to use WebSockets as the transport
                mechanism. Set to "tcp" to use raw TCP, which is the default.
            devices:
                List of device configurations that will be registered
                with the client. Consequently, the client will be able to
                subscribe to all registered device topics. Furthermore,
                after registration messages can simply published by the
                _devices id.
            service_groups:
                List of service group configurations that will be registered
                with the client. These should be known upon subscribing
                because the client will check for a matching service group if
                this is not known or registered with the IoT-Agent service
                the receiving of commands will fail. Please check the
                official documentation of the IoT-Agents API for more details.
            custom_encoder:
                Custom encoder class that will automatically parse the supported
                payload formats to a dictionary and vice versa. This
                essentially saves boiler plate code.
        """
        # initialize parent client
        super().__init__(
            client_id=client_id,
            clean_session=clean_session,
            userdata=userdata,
            protocol=protocol,
            callback_api_version=callback_api_version,
            transport=transport,
        )

        # setup logging functionality
        self.logger = logging.getLogger(name=f"{self.__class__.__name__}")
        self.logger.addHandler(logging.NullHandler())
        self.enable_logger(self.logger)

        # create dictionary holding the registered service groups
        self.service_groups: Dict[Tuple[str, str], ServiceGroup]
        if service_groups:
            self.service_groups = {gr.apikey: gr for gr in service_groups}
        else:
            self.service_groups = {}

        # create dictionary holding the registered device configurations
        # check if all _devices have the right transport protocol
        self._devices: Dict[str, Device] = {}
        if devices:
            self.devices = devices

        # create dict with available encoders
        self._encoders = {"IoTA-JSON": Json(), "PDI-IoTA-UltraLight": Ultralight()}

        # add custom encoder for message parsing
        if custom_encoder:
            self.add_encoder(custom_encoder)

    @property
    def devices(self):
        """
        Returns as list of all registered device configurations
        Returns:

        """
        return list(self._devices.values())

    @devices.setter
    def devices(self, devices: List[Device]):
        """
        Sets list of device configurations

        Args:
            devices: List of device configurations

        Returns:
            None

        Raises:
            ValueError: if duplicate device id was found
        """
        for device in devices:
            try:
                self.add_device(device=device)
            except ValueError:
                raise ValueError(f"Duplicate device_id: {device.device_id}")

    def get_encoder(self, encoder: Union[str, PayloadProtocol]):
        """
        Returns the encoder by key

        Args:
            encoder: encoder name

        Returns:
            Subclass of Baseencoder
        """
        return self._encoders.get(encoder)

    def add_encoder(self, encoder: Dict[str, BaseEncoder]):
        for value in encoder.values():
            assert isinstance(
                value, BaseEncoder
            ), f"Encoder must be a subclass of {type(BaseEncoder)}"

        self._encoders.update(encoder)

    def __validate_device(self, device: Union[Device, Dict]) -> Device:
        """
        Validates configuration of an IoT Device

        Args:
            device: device model to check on

        Returns:
            Device: validated model

        Raises:
            AssertionError: for faulty configurations
        """
        if isinstance(device, dict):
            device = Device.model_validate(device)

        assert isinstance(device, Device), "Invalid device configuration!"

        assert (
            device.transport == TransportProtocol.MQTT
        ), "Unsupported transport protocol found in device configuration!"

        if device.apikey in self.service_groups.keys():
            pass
        # check if matching service group is registered
        else:
            msg = (
                "Could not find matching service group! "
                "Communication may not work correctly!"
            )
            self.logger.warning(msg=msg)
            warnings.warn(message=msg)

        return device

    def __create_topic(
        self, *, topic_type: IoTAMQTTMessageType, device: Device, attribute: str = None
    ) -> str:
        """
        Creates a topic for a device configuration based on the requested
        topic type.

        Args:
            device:
                Configuration of an IoT device
            topic_type:
                type of the topic to be created,
                'multi' for topics that the device is suppose to publish on.
                'single' for topics that the device is suppose to publish on.
                'cmd' for topic the device is expecting its commands on.
                'cmdexe' for topic the device can acknowledge its commands on.
                'configuration' for topic the device can request command
                    configurations on
            attribute:
                attribute needs to be set for single measurements
        Returns:
            string with topic

        Raises:
            KeyError:
                If unknown message type is used
            ValueError:
                If attribute name is missing for single measurements
        """
        if topic_type == IoTAMQTTMessageType.MULTI:
            topic = "/".join(
                (
                    self._encoders[device.protocol].prefix,
                    device.apikey,
                    device.device_id,
                    "attrs",
                )
            )
        elif topic_type == IoTAMQTTMessageType.SINGLE:
            if attribute:
                attr = next(
                    attr for attr in device.attributes if attr.name == attribute
                )
                if attr.object_id:
                    attr_suffix = attr.object_id
                else:
                    attr_suffix = attr.name
                topic = "/".join(
                    (
                        self._encoders[device.protocol].prefix,
                        device.apikey,
                        device.device_id,
                        "attrs",
                        attr_suffix,
                    )
                )
            else:
                raise ValueError("Missing argument name for single measurement")
        elif topic_type == IoTAMQTTMessageType.CMD:
            topic = "/" + "/".join((device.apikey, device.device_id, "cmd"))
        elif topic_type == IoTAMQTTMessageType.CMDEXE:
            topic = "/".join(
                (
                    self._encoders[device.protocol].prefix,
                    device.apikey,
                    device.device_id,
                    "cmdexe",
                )
            )
        elif topic_type == IoTAMQTTMessageType.CONFIG:
            topic = "/".join(
                (
                    self._encoders[device.protocol].prefix,
                    device.apikey,
                    device.device_id,
                    "configuration",
                )
            )
        else:
            raise KeyError("topic_type not supported")
        return topic

    def __subscribe_commands(
        self, *, device: Device = None, qos=0, options=None, properties=None
    ):
        """
        Subscribes commands based on device configuration. If device argument is
        omitted the function will subscribe to all topics of already registered
        _devices.
        Additionally, it will also check if a matching service group is
        registered with the client. If nor a warning will be raised.

        Args:
            device: Configuration of an IoT device
            qos: Quality of service can be 0, 1 or 2
            options: MQTT v5.0 subscribe options
            properties: MQTT v5.0 properties

        Returns:
            None
        """
        if Device:
            if len(device.commands) > 0:
                topic = self.__create_topic(
                    device=device, topic_type=IoTAMQTTMessageType.CMD
                )
                super().subscribe(
                    topic=topic, qos=qos, options=options, properties=properties
                )
        else:
            # call itself but with device argument for all registered _devices
            for device in self._devices.values():
                self.__subscribe_commands(
                    device=device, qos=qos, options=options, properties=properties
                )

    def get_service_group(self, apikey: str) -> ServiceGroup:
        """
        Returns registered service group configuration

        Args:
            apikey: Unique APIKey of the service group

        Returns:
            ServiceGroup

        Raises:
            KeyError: if service group not yet registered

        Example::

            from filip.clients.mqtt import MQTTClient

            mqttc = MQTTClient()
            group = mqttc.get_service_group(apikey="MyAPIKEY")
            print(group.json(indent=2))
            print(type(group))
        """
        group = self.service_groups.get(apikey, None)
        if group is None:
            raise KeyError("Service group with apikey %s not found!", apikey)
        return group

    def add_service_group(self, service_group: Union[ServiceGroup, Dict]):
        """
        Registers a device service group with the client

        Args:
            service_group: Service group configuration

        Returns:
            None

        Raises:
            ValueError: if service group already exists
        """
        if isinstance(service_group, dict):
            service_group = ServiceGroup.model_validate(service_group)
        assert isinstance(
            service_group, ServiceGroup
        ), "Invalid content for service group!"

        if self.service_groups.get(service_group.apikey, None) is None:
            pass
        else:
            raise ValueError("Service group already exists! %s", service_group.apikey)
        # add service group configuration to the service group list
        self.service_groups[service_group.apikey] = service_group

    def delete_service_group(self, apikey):
        """
        Unregisters a service group and removes

        Args:
            apikey: Unique APIKey of the service group

        Returns:
            None
        """
        group = self.service_groups.pop(apikey, None)
        if group:
            self.logger.info("Successfully unregistered Service Group '%s'!", apikey)
        else:
            self.logger.error("Could not unregister service group '%s'!", apikey)
            raise KeyError("Device not found!")

    def update_service_group(self, service_group: Union[ServiceGroup, Dict]):
        """
        Updates a registered service group configuration. There is no
        opportunity to only partially update the device. Hence, your service
        group model should be complete.

        Args:
            service_group: Service group configuration

        Returns:
            None

        Raises:
            KeyError: if service group not yet registered
        """
        if isinstance(service_group, dict):
            service_group = ServiceGroup.model_validate(service_group)
        assert isinstance(
            service_group, ServiceGroup
        ), "Invalid content for service group"

        if self.service_groups.get(service_group.apikey, None) is None:
            raise KeyError("Service group not found! %s", service_group.apikey)
        # add service group configuration to the service group list
        self.service_groups[service_group.apikey] = service_group

    def get_device(self, device_id: str) -> Device:
        """
        Returns the configuration of a registered device.

        Args:
            device_id: Id of the requested device

        Returns:
           Device: Device model of the requested device

        Raises:
            KeyError: if requested device is not registered with the client

        Example::

            from filip.clients.mqtt import MQTTClient

            mqttc = MQTTClient()
            device = mqttc.get_device(device_id="MyDeviceId")
            print(device.json(indent=2))
            print(type(device))
        """
        return self._devices[device_id]

    def add_device(
        self, device: Union[Device, Dict], qos=0, options=None, properties=None
    ):
        """
        Registers a device config with the mqtt client. Subsequently,
        the client will magically subscribe to the corresponding topics based
        on the device config and any matching registered service group config
        if exists.

        Note:
            To register the device config only with this client is not
            sufficient for data streaming the configuration also needs to be
            registered with IoTA-Agent.

        Args:
            device: Configuration of an IoT device
            qos: Quality of service can be 0, 1 or 2
            options: MQTT v5.0 subscribe options
            properties: MQTT v5.0 properties

        Returns:
            None

        Raises:
            ValueError: if device configuration already exists
        """
        device = self.__validate_device(device=device)

        if self._devices.get(device.device_id, None) is None:
            pass
        else:
            raise ValueError("Device already exists! %s", device.device_id)
        # add device configuration to the device list
        self._devices[device.device_id] = device
        # subscribes to the command topic
        self.__subscribe_commands(
            device=device, qos=qos, options=options, properties=properties
        )

    def delete_device(self, device_id: str):
        """
        Unregisters a device and removes its subscriptions and callbacks

        Args:
            device_id: id of and IoT device

        Returns:
            None
        """
        device = self._devices.pop(device_id, None)
        if device:
            topic = self.__create_topic(
                device=device, topic_type=IoTAMQTTMessageType.CMD
            )
            self.unsubscribe(topic=topic)
            self.message_callback_remove(sub=topic)
            self.logger.info("Successfully unregistered Device '%s'!", device_id)
        else:
            self.logger.error("Could not unregister device '%s'", device_id)

    def update_device(
        self, device: Union[Device, Dict], qos=0, options=None, properties=None
    ):
        """
        Updates a registered device configuration. There is no opportunity
        to only partially update the device. Hence, your device model should
        be complete.

        Args:
            device: Configuration of an IoT device
            qos: Quality of service can be 0, 1 or 2
            options: MQTT v5.0 subscribe options
            properties: MQTT v5.0 properties

        Returns:
            None

        Raises:
            KeyError: if device not yet registered
        """
        device = self.__validate_device(device=device)

        if self._devices.get(device.device_id, None) is None:
            raise KeyError("Device not found! %s", device.device_id)

        # update device configuration in the device list
        self._devices[device.device_id] = device
        # subscribes to the command topic
        self.__subscribe_commands(
            device=device, qos=qos, options=options, properties=properties
        )

    def add_command_callback(self, device_id: str, callback: Callable):
        """
        Adds callback function for a device configuration.

        Args:
            device_id:
                id of and IoT device
            callback:
                function that will be called for incoming commands.
                This function should have the following format:

        Example::

            def on_command(client, obj, msg):
                apikey, device_id, payload = \
                    client.encoder.decode_message(msg=msg)

                # do_something with the message.
                # For instance write into a queue.

                # acknowledge a command. Here command are usually single
                # messages. The first key is equal to the commands name.
                client.publish(device_id=device_id,
                               command_name=next(iter(payload)),
                               payload=payload)

            mqttc.add_command_callback(device_id="MyDevice",
                                       callback=on_command)

        Returns:
            None
        """
        device = self._devices.get(device_id, None)
        if device is None:
            raise KeyError("Device does not exist! %s", device_id)
        self.__subscribe_commands(device=device)
        topic = self.__create_topic(device=device, topic_type=IoTAMQTTMessageType.CMD)
        self.message_callback_add(topic, callback)

    def publish(
        self,
        topic=None,
        payload: Union[Dict, Any] = None,
        qos: int = 0,
        retain: bool = False,
        properties=None,
        device_id: str = None,
        attribute_name: str = None,
        command_name: str = None,
        timestamp: bool = False,
    ):
        """
        Publish an MQTT Message to a specified topic. If you want to publish
        a device specific message to a device use the device_id argument for
        multi-measurement. The function will then automatically validate
        against the registered device configuration if the payload keys are
        valid. If you want to publish a single measurement the attribute_name
        argument is required as well.

        Note:
            If the device_id argument is set, the topic argument will be
            ignored.

        Args:
            topic:
                The topic that the message should be published on.
            payload:
                The actual message to send. If not given, or set to None a
                zero length message will be used. Passing an int or float will
                result in the payload being converted to a string
                representing that number. If you wish to send a true
                int/float, use struct.pack() to create the
                payload you require. For publishing to a device use a dict
                containing the object_ids as keys.
            qos:
                The quality of service level to use.
            retain:
                If set to true, the message will be set as the "last known
                good"/retained message for the topic.
            properties:
                (MQTT v5.0 only) the MQTT v5.0 properties to be included.
                Use the Properties class.
            device_id:
                Id of the IoT device you want to publish for. The topics will
                automatically created. If set, the message type will be
                assumed to be multi measurement.
            attribute_name:
                Name of an attribute of the device. Do only use this for
                single measurements. If set, `command_name` must
                be omitted.
            command_name:
                Name of a command of the device that should be acknowledged. If
                set `attribute_name` must be omitted.
            timestamp:
                If `true` the client will generate a valid timestamp based on
                utc and added to the multi measurement payload.
                If a `timeInstant` is already contained in the
                message payload it will not overwritten.

        Returns:
            None

        Raises:
            KeyError: if device configuration is not registered with client
            ValueError: if the passed arguments are inconsistent or a
                timestamp does not match the ISO 8601 format.
            AssertionError: if the message payload does not match the device
                configuration.
        """

        # TODO: time stamps are not tested yet

        if device_id:
            device = self.get_device(device_id=device_id)

            # create message for multi measurement payload
            if attribute_name is None and command_name is None:
                assert isinstance(payload, dict), "Payload must be a dictionary"

                if timestamp and "timeInstant" not in payload.keys():
                    payload["timeInstant"] = datetime.utcnow()
                # validate if dict keys match device configuration

                msg_payload = payload.copy()
                for key in payload.keys():
                    for attr in device.attributes:
                        key_constraint = key == "timeInstant"

                        def elif_action(msg):
                            None

                        if attr.object_id is not None:
                            key_constraint = key_constraint or (key in attr.object_id)

                            def elif_action(msg):
                                msg[attr.object_id] = msg.pop(key)

                        # could be made more compact by pulling up the second condition
                        # but would probably make the code less readable...
                        if key_constraint:
                            break

                        elif key == attr.name:
                            elif_action(msg_payload)
                            break

                    else:
                        err_msg = (
                            f"Attribute key '{key}' is not allowed "
                            f"in the message payload for this "
                            f"device configuration with device_id "
                            f"'{device_id}'"
                        )
                        raise KeyError(err_msg)
                topic = self.__create_topic(
                    device=device, topic_type=IoTAMQTTMessageType.MULTI
                )
                payload = self._encoders[device.protocol].encode_msg(
                    device_id=device_id,
                    payload=msg_payload,
                    msg_type=IoTAMQTTMessageType.MULTI,
                )

            # create message for command acknowledgement
            elif attribute_name is None and command_name:
                assert isinstance(payload, Dict), "Payload must be a dictionary"
                assert (
                    len(payload.keys()) == 1
                ), "Cannot acknowledge multiple commands simultaneously"
                assert next(iter(payload.keys())) in [
                    cmd.name for cmd in device.commands
                ], "Unknown command for this device!"
                topic = self.__create_topic(
                    device=device, topic_type=IoTAMQTTMessageType.CMDEXE
                )
                payload = self._encoders[device.protocol].encode_msg(
                    device_id=device_id,
                    payload=payload,
                    msg_type=IoTAMQTTMessageType.CMDEXE,
                )

            # create message for single measurement
            elif attribute_name and command_name is None:
                topic = self.__create_topic(
                    device=device,
                    topic_type=IoTAMQTTMessageType.SINGLE,
                    attribute=attribute_name,
                )
                payload = self._encoders[device.protocol].encode_msg(
                    device_id=device_id,
                    payload=payload,
                    msg_type=IoTAMQTTMessageType.SINGLE,
                )
            else:
                raise ValueError("Inconsistent arguments!")

        super().publish(
            topic=topic, payload=payload, qos=qos, retain=retain, properties=properties
        )

    def subscribe(self, topic=None, qos=0, options=None, properties=None):
        """
        Extends the normal subscribe function of the paho.mqtt.client.
        If the topic argument is omitted the client will subscribe to all
        registered device command topics.

        Args:
            topic:
                A string specifying the subscription topic to subscribe to.
            qos:
                The desired quality of service level for the subscription.
                Defaults to 0.
            options: Not used.
            properties: Not used.

        Returns:
            None
        """
        if topic:
            super().subscribe(
                topic=topic, qos=qos, options=options, properties=properties
            )
        else:
            for device in self._devices.values():
                self.__subscribe_commands(
                    device=device, qos=qos, options=options, properties=properties
                )



"""
Module for FIWARE api client
"""

import logging
import json
import errno
from typing import Optional, Union, Dict
from pathlib import Path
from pydantic import BaseModel, AnyHttpUrl
from requests.auth import HTTPBasicAuth, HTTPDigestAuth
from requests import Session
from filip.clients.base_http_client import BaseHttpClient
from filip.config import settings
from filip.models.base import FiwareHeader
from filip.clients.ngsi_v2 import ContextBrokerClient, IoTAClient, QuantumLeapClient


logger = logging.getLogger("client")


class HttpClientConfig(BaseModel):
    """
    Config class for http client
    """

    cb_url: Optional[AnyHttpUrl] = settings.CB_URL
    iota_url: Optional[AnyHttpUrl] = settings.IOTA_URL
    ql_url: Optional[AnyHttpUrl] = settings.QL_URL
    auth: Optional[Dict] = None


class HttpClient(BaseHttpClient):
    """
    Master client. This client contains all implemented sub clients based on
    the principal of composition. Hence, each sub client is accessible from
    this client, but they share a general config and if provided a session.
    """

    def __init__(
        self,
        config: Union[str, Path, HttpClientConfig, Dict] = None,
        session: Session = None,
        fiware_header: FiwareHeader = None,
        **kwargs
    ):
        """
        Constructor for master client
        Args:
            config (Union[str, Path, Dict]): Configuration object
            session (request.Session): Session object
            fiware_header (FiwareHeader): Fiware header
            **kwargs: Optional arguments that ``request`` takes.
        """
        if config:
            self.config = config
        else:
            self.config = HttpClientConfig()

        super().__init__(session=session, fiware_header=fiware_header, **kwargs)

        # initialize sub clients
        self.cb = ContextBrokerClient(
            url=self.config.cb_url,
            session=self.session,
            fiware_header=self.fiware_headers,
            **self.kwargs
        )

        self.iota = IoTAClient(
            url=self.config.iota_url,
            session=self.session,
            fiware_header=self.fiware_headers,
            **self.kwargs
        )

        self.timeseries = QuantumLeapClient(
            url=self.config.ql_url,
            session=self.session,
            fiware_header=self.fiware_headers,
            **self.kwargs
        )

        # from here on deprecated?
        auth_types = {
            "basicauth": self.__http_basic_auth,
            "digestauth": self.__http_digest_auth,
        }
        # 'oauth2': self.__oauth2}

        if self.config.auth:
            assert self.config.auth["type"].lower() in auth_types.keys()
            self.__get_secrets_file(path=self.config.auth["secret"])
            auth_types[self.config.auth["type"]]()

        self.__secrets = {
            "username": None,
            "password": None,
            "client_id": None,
            "client_secret": None,
        }

    @property
    def config(self):
        """Return current config"""
        return self._config

    @config.setter
    def config(self, config: HttpClientConfig):
        """Set a new config"""
        if isinstance(config, HttpClientConfig):
            self._config = config
        elif isinstance(config, (str, Path)):
            with open(config) as f:
                config_json = f.read()
                self._config = HttpClientConfig.model_validate_json(config_json)
        else:
            self._config = HttpClientConfig.model_validate(config)

    @property
    def cert(self):
        """Return session certificate"""
        return self.session.cert

    @property
    def secrets(self):
        """Returns secrets"""
        return self.__secrets

    @secrets.setter
    def secrets(self, data: dict):
        """Set new secrets"""
        self.__secrets.update(data)

    @secrets.deleter
    def secrets(self):
        """Delete secrets"""
        self.__secrets = {}

    def __get_secrets_file(self, path=None):
        """
        Reads credentials form secret file the path variable is pointing to.

        Args:
            path: location of secrets-file
        Returns:
             None
        """
        try:
            with open(path, "r") as filename:
                logger.info("Reading credentials from: %s", path)
                self.__secrets.update(json.load(filename))

        except IOError as err:
            if err.errno == errno.ENOENT:
                logger.error("%s - does not exist", path)
            elif err.errno == errno.EACCES:
                logger.error("%s - cannot be read", path)
            else:
                logger.error("%s - some other error", path)

    def __http_basic_auth(self):
        """
        Initiates a client using the basic authorization mechanism provided by
        the requests package. The documentation of the package is located here:
        https://requests.readthedocs.io/en/master/user/authentication/
        The credentials must be provided via secret-file.
        """
        try:
            self.session = Session()
            self.session.auth = HTTPBasicAuth(
                self.__secrets["username"], self.__secrets["password"]
            )
        except KeyError:
            pass

    def __http_digest_auth(self):
        """
        Initiates a client using the digest authorization mechanism provided by
        the requests package. The documentation of the package is located here:
        https://requests.readthedocs.io/en/master/user/authentication/
        The credentials must be provided via secret-file.
        """
        try:
            self.session = Session()
            self.session.auth = HTTPDigestAuth(
                self.__secrets["username"], self.__secrets["password"]
            )
        except KeyError:
            pass

    # def __oauth2(self):
    #    """
    #    Initiates a oauthclient according to the workflows defined by OAuth2.0.
    #    We use requests-oauthlib for this implementation. The documentation
    #    of the package is located here:
    #    https://requests-oauthlib.readthedocs.io/en/latest/index.html
    #    The information for workflow selection must be provided via
    #    filip-config. The credentials must be provided via secrets-file.
    #    :return: None
    #    """
    #    oauth2clients = {'authorization_code': None,
    #                     'implicit': MobileApplicationClient,
    #                     'resource_owner_password_credentials':
    #                         LegacyApplicationClient,
    #                     'client_credentials': BackendApplicationClient, }
    #    try:
    #        workflow = self.config['auth']['workflow']
    #    except KeyError:
    #        logger.warning(f"No workflow for OAuth2 defined! Default "
    #                       f"workflow will used: Authorization Code Grant."
    #                       f"Other oauth2-workflows available are: "
    #                       f"{oauth2clients.keys()}")
    #        workflow = 'authorization_code_grant'
    #
    #    oauthclient = oauth2clients[workflow](client_id=self.__secrets[
    #        'client_id'])
    #    self.session = OAuth2Session(client_id=None,
    #                                 client=oauthclient,
    #                                 auto_refresh_url=self.__secrets[
    #                                     'token_url'],
    #                                 auto_refresh_kwargs={
    #                                     self.__secrets['client_id'],
    #                                     self.__secrets['client_secret']})
    #
    #    self.__token = self.session.fetch_token(
    #        token_url=self.__secrets['token_url'],
    #        username=self.__secrets['username'],
    #        password=self.__secrets['password'],
    #        client_id=self.__secrets['client_id'],
    #        client_secret=self.__secrets['client_secret'])

    def __token_saver(self, token):
        self.__token = token



"""
The PostProcessing gets called after the vocabulary was parsed from sources

The postprocessing has the goal to add predefined values,
compute combinedRelations, reload user settings, and precompute
information as: duplicate labels or sort relations
"""

import datetime
import re
from typing import List, Optional

import stringcase

from filip.semantics.ontology_parser.vocabulary_builder import VocabularyBuilder
from filip.semantics.vocabulary import (
    Source,
    IdType,
    Vocabulary,
    DatatypeType,
    Datatype,
    Class,
)
from filip.semantics.vocabulary import (
    CombinedDataRelation,
    CombinedObjectRelation,
    CombinedRelation,
)


class PostProcessor:
    """Class offering postprocessing as cls-methods for a vocabulary"""

    @classmethod
    def post_process_vocabulary(
        cls, vocabulary: Vocabulary, old_vocabulary: Optional[Vocabulary] = None
    ):
        """Main methode to be called for post processing

        Args:
            vocabulary (Vocabulary): Freshly parsed Vocabulary
            old_vocabulary (Vocabulary): Existing Vocabulary of which the
                settings should be overtaken

        Returns:
            None
        """

        # all methods have to reset the state that they are editing first.
        # consecutive calls of post_process_vocabulary need to have the same
        # result
        voc_builder = VocabularyBuilder(vocabulary=vocabulary)
        cls._set_labels(voc_builder)
        cls._add_predefined_source(voc_builder)
        cls._add_predefined_datatypes(voc_builder)
        cls._add_owl_thing(voc_builder)
        cls._remove_duplicate_parents(voc_builder)

        cls._log_and_clear_dependencies(voc_builder)
        cls._compute_ancestor_classes(voc_builder)
        cls._compute_child_classes(voc_builder)
        cls._combine_relations(voc_builder)

        if old_vocabulary is not None:
            cls.transfer_settings(
                new_vocabulary=vocabulary, old_vocabulary=old_vocabulary
            )
        cls._apply_vocabulary_settings(voc_builder)

        cls._ensure_parent_class(voc_builder)

        cls._sort_relations(voc_builder)
        cls._mirror_object_property_inverses(voc_builder)

        cls._save_initial_label_summary(vocabulary)

    @classmethod
    def _set_labels(cls, voc_builder: VocabularyBuilder):
        """If entities have no label, extract their label from the iri

        Args:
            voc_builder: Builder object for Vocabulary

        Returns:
            None
        """
        for entity in voc_builder.vocabulary.get_all_entities():
            entity.label = entity.get_original_label()

    @classmethod
    def _add_predefined_source(cls, voc_builder: VocabularyBuilder):
        """Add a special source to the vocabulary: PREDEFINED

        Args:
            voc_builder: Builder object for Vocabulary

        Returns:
            None
        """
        if "PREDEFINED" not in voc_builder.vocabulary.sources:
            source = Source(
                source_name="Predefined",
                timestamp=datetime.datetime.now(),
                predefined=True,
            )
            voc_builder.add_source(source, "PREDEFINED")

    @classmethod
    def _log_and_clear_dependencies(cls, voc_builder: VocabularyBuilder):
        """
        remove all references to entities that are not in the vocabulary to
        prevent program errrors as we remove information we need to reparse
        the source each time a new source is added as than the dependency
        could be valid. Further log the found dependencies for the user to
        display

        Args:
            voc_builder: Builder object for Vocabulary

        Returns:
            None
        """
        for ontology in voc_builder.vocabulary.sources.values():
            ontology.treat_dependency_statements(voc_builder.vocabulary)

    @classmethod
    def _add_predefined_datatypes(cls, voc_builder: VocabularyBuilder):
        """
        Add predefinded datatype_catalogue to the PREDEFINED source; they
        are not included in an OWL file

        Args:
            voc_builder: Builder object for Vocabulary

        Returns:
            None
        """
        # Test if datatype_catalogue were already added, if yes skip
        if (
            "http://www.w3.org/2002/07/owl#rational"
            in voc_builder.vocabulary.datatypes.keys()
        ):
            return

        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2002/07/owl#rational",
                comment="All numbers allowed",
                type=DatatypeType.number,
                number_decimal_allowed=True,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2002/07/owl#real",
                comment="All whole numbers allowed",
                type=DatatypeType.number,
                number_decimal_allowed=False,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/1999/02/22-rdf-syntax-ns#PlainLiteral",
                comment="All strings allowed",
                type=DatatypeType.string,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/1999/02/22-rdf-syntax-ns#XMLLiteral",
                comment="XML Syntax required",
                type=DatatypeType.string,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2000/01/rdf-schema#Literal",
                comment="All strings allowed",
                type=DatatypeType.string,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#anyURI",
                comment="Needs to start with http://",
                type=DatatypeType.string,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#base64Binary",
                comment="Base64Binary",
                type=DatatypeType.string,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#boolean",
                comment="True or False",
                type=DatatypeType.enum,
                enum_values=["True", "False"],
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#byte",
                comment="Byte Number",
                type=DatatypeType.number,
                number_has_range=True,
                number_range_min=-128,
                number_range_max=127,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#dateTime",
                comment="Date with possible timezone",
                type=DatatypeType.date,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#dateTimeStamp",
                comment="Date",
                type=DatatypeType.date,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#decimal",
                comment="All decimal numbers",
                type=DatatypeType.number,
                number_decimal_allowed=True,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#double",
                comment="64 bit decimal",
                type=DatatypeType.number,
                number_decimal_allowed=True,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#float",
                comment="32 bit decimal",
                type=DatatypeType.number,
                number_decimal_allowed=True,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#hexBinary",
                comment="Hexadecimal",
                type=DatatypeType.string,
                allowed_chars=[
                    "0",
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "A",
                    "B",
                    "C",
                    "D",
                    "E",
                    "F",
                ],
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#int",
                comment="Signed 32 bit number",
                type=DatatypeType.number,
                number_has_range=True,
                number_range_min=-2147483648,
                number_range_max=2147483647,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#integer",
                comment="All whole numbers",
                type=DatatypeType.number,
                number_decimal_allowed=False,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#language",
                comment="Language code, e.g: en, en-US, fr, or fr-FR",
                type=DatatypeType.string,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#long",
                comment="Signed 64 bit integer",
                type=DatatypeType.number,
                number_has_range=True,
                number_range_min=-9223372036854775808,
                number_range_max=9223372036854775807,
                number_decimal_allowed=False,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#Name",
                comment="Name string (dont start with number)",
                type=DatatypeType.string,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#NCName",
                comment="Name string : forbidden",
                type=DatatypeType.string,
                forbidden_chars=[":"],
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#negativeInteger",
                comment="All negative whole numbers",
                type=DatatypeType.number,
                number_has_range=True,
                number_range_max=-1,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#NMTOKEN",
                comment="Token string",
                type=DatatypeType.string,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#nonNegativeInteger",
                comment="All positive whole numbers",
                type=DatatypeType.number,
                number_has_range=True,
                number_range_min=0,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#nonPositiveInteger",
                comment="All negative whole numbers",
                type=DatatypeType.number,
                number_has_range=True,
                number_range_max=-1,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#normalizedString",
                comment="normalized String",
                type=DatatypeType.string,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#positiveInteger",
                comment="All positive whole numbers",
                type=DatatypeType.number,
                number_has_range=True,
                number_range_min=0,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#short",
                comment="signed 16 bit number",
                type=DatatypeType.number,
                number_has_range=True,
                number_range_min=-32768,
                number_range_max=32767,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#string",
                comment="String",
                type=DatatypeType.string,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#token",
                comment="String",
                type=DatatypeType.string,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#unsignedByte",
                comment="unsigned 8 bit number",
                type=DatatypeType.number,
                number_has_range=True,
                number_range_min=0,
                number_range_max=255,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#unsignedInt",
                comment="unsigned 32 bit number",
                type=DatatypeType.number,
                number_has_range=True,
                number_range_min=0,
                number_range_max=4294967295,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#unsignedLong",
                comment="unsigned 64 bit number",
                type=DatatypeType.number,
                number_has_range=True,
                number_range_min=0,
                number_range_max=18446744073709551615,
            )
        )
        voc_builder.add_predefined_datatype(
            Datatype(
                iri="http://www.w3.org/2001/XMLSchema#unsignedShort",
                comment="unsigned 16 bit number",
                type=DatatypeType.number,
                number_has_range=True,
                number_range_min=0,
                number_range_max=65535,
            )
        )

    @classmethod
    def _add_owl_thing(cls, voc_builder: VocabularyBuilder):
        """Add owl_thing class to the vocabulary in the predefined source

        By definition each class is a subclass of owl:thing and owl:thing can be
        a target of relation but owl thing is never mentioned explicitly in
        ontology files.

        Args:
            voc_builder: Builder object for Vocabulary
        Returns:
            None
        """
        root_class = Class(
            iri="http://www.w3.org/2002/07/owl#Thing",
            comment="Predefined root_class",
            label="Thing",
            predefined=True,
        )

        # as it is the root object it is only a parent of classes which have no
        # parents yet
        for class_ in voc_builder.vocabulary.get_classes():
            if class_.parent_class_iris == []:
                class_.parent_class_iris.insert(0, root_class.iri)

        if root_class.iri not in voc_builder.vocabulary.classes:
            voc_builder.add_class(root_class)
            root_class.source_ids.add("PREDEFINED")

    @classmethod
    def _remove_duplicate_parents(cls, voc_builder: VocabularyBuilder):
        """Prevent that a class_ has the same parent iri multiple times

        Args:
            voc_builder: Builder object for Vocabulary
        Returns:
            None
        """
        for class_ in voc_builder.vocabulary.classes.values():
            class_.parent_class_iris = list(dict.fromkeys(class_.parent_class_iris))

    @classmethod
    def _ensure_parent_class(cls, voc_builder: VocabularyBuilder):
        """If a class has a parent class, which was provided by an other
        ontology. And that ontology is not given, it will have no parents.
        In that case give him Thing as direct parent

        Args:
            voc_builder: Builder object for Vocabulary
        Returns:
            None
        """
        for class_ in voc_builder.vocabulary.classes.values():
            # Thing is the root of all
            if not class_.iri == "http://www.w3.org/2002/07/owl#Thing":
                if len(class_.parent_class_iris) == 0:
                    class_.parent_class_iris.append(
                        "http://www.w3.org/2002/07/owl#Thing"
                    )

    @classmethod
    def _apply_vocabulary_settings(cls, voc_builder: VocabularyBuilder):
        """
        Make the labels of all entities FIWARE safe, so that they can be used
        as field keys

        Args:
            voc_builder: Builder object for Vocabulary
        Returns:
            None
        """
        vocabulary = voc_builder.vocabulary
        settings = vocabulary.settings

        def to_pascal_case(string: str) -> str:
            return (
                stringcase.pascalcase(string)
                .replace("_", "")
                .replace(" ", "")
                .replace("-", "")
            )

        def to_camel_case(string: str) -> str:
            camel_string = stringcase.camelcase(string)
            return camel_string

        def to_snake_case(string: str) -> str:
            camel_string = to_pascal_case(string)
            return re.sub(r"(?<!^)(?=[A-Z])", "_", camel_string).lower()

        # replace all whitespaces
        for entity in vocabulary.get_all_entities():
            entity.label = entity.label.replace(" ", "_")

        # replace al whitespaces in enum_values
        for datatype in vocabulary.datatypes.values():
            new_enums = []
            for enum in datatype.enum_values:
                new_enums.append(enum.replace(" ", "_"))
            datatype.enum_values = new_enums

        if settings.pascal_case_class_labels:
            for class_ in vocabulary.get_classes():
                class_.label = to_pascal_case(class_.label)

        if settings.pascal_case_individual_labels:
            for individual in vocabulary.individuals.values():
                individual.label = to_pascal_case(individual.label)

        if settings.camel_case_property_labels:
            props = list(vocabulary.data_properties.values())
            props.extend(vocabulary.object_properties.values())
            for prop in props:
                prop.label = to_camel_case(prop.label)

        if settings.camel_case_datatype_labels:
            for datatype in vocabulary.datatypes.values():
                datatype.label = to_camel_case(datatype.label)

        if settings.pascal_case_datatype_enum_labels:
            for datatype in vocabulary.get_enum_dataytypes().values():
                datatype.label = to_pascal_case(datatype.label)

    @classmethod
    def _save_initial_label_summary(cls, vocabulary: Vocabulary):
        """
        Save the label_summary existing after parsing, before the user
        changed labels

        Args:
            vocabulary: vocabulary of which the label summary should be saved

        Returns:
            None
        """
        from filip.semantics.vocabulary_configurator import VocabularyConfigurator

        vocabulary.original_label_summary = (
            VocabularyConfigurator.get_label_conflicts_in_vocabulary(
                vocabulary=vocabulary
            )
        )

    @classmethod
    def _compute_ancestor_classes(cls, voc_builder: VocabularyBuilder):
        """Compute all ancestor classes of classes

        Args:
            voc_builder: Builder object for Vocabulary
        Returns:
            None
        """
        vocabulary = voc_builder.vocabulary
        # clear state
        for class_ in vocabulary.get_classes():
            class_.ancestor_class_iris = []

        for class_ in vocabulary.get_classes():
            queue: List[str] = []
            queue.extend(class_.parent_class_iris)

            while len(queue) > 0:
                parent = queue.pop()

                if not voc_builder.entity_is_known(parent):
                    continue

                class_.ancestor_class_iris.append(parent)
                grand_parents = vocabulary.get_class_by_iri(parent).parent_class_iris

                for grand_parent in grand_parents:
                    if grand_parent not in class_.ancestor_class_iris:
                        # prevent infinite loop if inheritance circle
                        queue.append(grand_parent)

    @classmethod
    def _compute_child_classes(cls, voc_builder: VocabularyBuilder):
        """Compute all child classes of classes

        Args:
            voc_builder: Builder object for Vocabulary
        Returns:
            None
        """
        vocabulary = voc_builder.vocabulary
        # clear state
        for class_ in vocabulary.get_classes():
            class_.child_class_iris = []

        for class_ in vocabulary.get_classes():
            for parent in class_.ancestor_class_iris:

                if not voc_builder.entity_is_known(parent):
                    continue

                parent_class = vocabulary.get_class_by_iri(parent)
                parent_class.child_class_iris.append(class_.iri)

    @classmethod
    def _combine_relations(cls, voc_builder: VocabularyBuilder):
        """Compute all CombinedRelations

        Args:
            voc_builder: Builder object for Vocabulary
        Returns:
            None
        """
        vocabulary = voc_builder.vocabulary
        # clear state
        vocabulary.combined_object_relations.clear()
        vocabulary.combined_data_relations.clear()

        for class_ in vocabulary.get_classes():
            class_.combined_object_relation_ids = []
            class_.combined_data_relation_ids = []

        for class_ in vocabulary.get_classes():

            relations_with_property_iri = {}

            all_relation_ids = []
            all_relation_ids.extend(class_.get_relation_ids())
            for ancestor_iri in class_.ancestor_class_iris:

                if not voc_builder.entity_is_known(ancestor_iri):
                    continue
                ancestor = vocabulary.get_class_by_iri(ancestor_iri)
                all_relation_ids.extend(ancestor.get_relation_ids())

            for relation_id in all_relation_ids:
                relation = vocabulary.get_relation_by_id(id=relation_id)
                property_iri = relation.property_iri

                if property_iri not in relations_with_property_iri:
                    relations_with_property_iri[property_iri] = []

                relations_with_property_iri[property_iri].append(relation_id)

            for property_iri, rel_list in relations_with_property_iri.items():

                # These ids should be derived, so that the same combined
                # relation always ends up with the same id as a class can
                # only have 1 combined relation of a property these ids are
                # unique by keeping the ids always the same, we can store
                # information more efficiently in the database (settings)

                # if a property iri is not known while parsing an ontology
                # (dependency not yet parsed) the relations with that
                # property are going to get ignored, maybe a not should be
                # displayed
                if vocabulary.is_id_of_type(property_iri, IdType.data_property):
                    id = "combined-data-relation|{}|{}".format(class_.iri, property_iri)
                    combi = CombinedDataRelation(
                        id=id,
                        property_iri=property_iri,
                        relation_ids=rel_list,
                        class_iri=class_.iri,
                    )
                    voc_builder.add_combined_data_relation_for_class(
                        class_iri=class_.iri, cdata=combi
                    )
                elif vocabulary.is_id_of_type(property_iri, IdType.object_property):
                    id = "combined-object-relation|{}|{}".format(
                        class_.iri, property_iri
                    )
                    combi = CombinedObjectRelation(
                        id=id,
                        property_iri=property_iri,
                        relation_ids=rel_list,
                        class_iri=class_.iri,
                    )
                    voc_builder.add_combined_object_relation_for_class(
                        class_iri=class_.iri, crel=combi
                    )
                else:
                    pass

    @classmethod
    def _sort_relations(cls, voc_builder: VocabularyBuilder):
        """sort relations alphabetically according to their labels

        Args:
            voc_builder: Builder object for Vocabulary
        Returns:
            None
        """
        vocabulary = voc_builder.vocabulary

        for class_ in vocabulary.get_classes():
            cors = class_.get_combined_object_relations(vocabulary)
            class_.combined_object_relation_ids = cls._sort_list_of_combined_relations(
                cors, vocabulary
            )

            cdrs = class_.get_combined_data_relations(vocabulary)
            class_.combined_data_relation_ids = cls._sort_list_of_combined_relations(
                cdrs, vocabulary
            )

    @classmethod
    def _sort_list_of_combined_relations(
        cls, combined_relations: List[CombinedRelation], vocabulary: Vocabulary
    ) -> List[str]:
        """sort given CombinedRelations according to their labels

        Args:
            vocabulary (Vocabulary)
            combined_relations (List[CombinedRelation]): CRs to sort
        Returns:
            List[str], list of cr_id, sorted according to their label
        """

        property_dic = {}

        for cor in combined_relations:
            property_iri = cor.property_iri
            label = cor.get_property_label(vocabulary=vocabulary)
            property_dic[label + property_iri] = cor.id
            # combine label with iri to prevent an error due to two identical
            # labels
        sorted_property_dic = sorted(property_dic.items())

        sorted_cor_ids = []
        for pair in sorted_property_dic:
            sorted_cor_ids.append(pair[1])
        return sorted_cor_ids

    @classmethod
    def _mirror_object_property_inverses(cls, voc_builder: VocabularyBuilder):
        """
        inverses could only be given for 1 obj_prop of the pair and needs to
        be derived for the other also we could have the inverse inside an other
        import (there for done in postprocessing)

        Args:
            voc_builder: Builder object for Vocabulary

        Returns:
            None
        """
        # the state is not cleared, instead add_inverse_property_iri() makes
        # sure that there will be no duplicates as it is a set
        vocabulary = voc_builder.vocabulary

        for obj_prop_iri in vocabulary.object_properties:
            obj_prop = vocabulary.get_object_property(obj_prop_iri)

            for inverse_iri in obj_prop.inverse_property_iris:
                inverse_prop = vocabulary.get_object_property(inverse_iri)
                inverse_prop.add_inverse_property_iri(obj_prop_iri)

    @classmethod
    def transfer_settings(cls, new_vocabulary: Vocabulary, old_vocabulary: Vocabulary):
        """
        Transfer all the user made settings (labels, ..)
        from an old vocabulary to a new vocabulary

        Args:
            new_vocabulary (Vocabulary): Vocabulary to which the settings should
                be transferred
            old_vocabulary (Vocabulary): Vocabulary of which the settings should
                be transferred

        Returns:
            None
        """

        # label settings
        for entity in old_vocabulary.get_all_entities():
            new_entity = new_vocabulary.get_entity_by_iri(entity.iri)
            if new_entity is not None:
                new_entity.user_set_label = entity.user_set_label

        # device settings
        for iri, data_property in old_vocabulary.data_properties.items():
            if iri in new_vocabulary.data_properties:
                new_data_property = new_vocabulary.data_properties[iri]
                new_data_property.field_type = data_property.field_type



from pydantic import validate_arguments
from typing import Any, Dict, Tuple, Union
from filip.clients.mqtt.encoder import BaseEncoder
from filip.models.mqtt import IoTAMQTTMessageType


class Ultralight(BaseEncoder):
    prefix = "/ul"

    def __init__(self):
        super().__init__()

    @staticmethod
    @validate_arguments
    def __eval_value(value: Union[bool, float, str]):
        return value

    def decode_message(self, msg, decoder="utf-8") -> Tuple[str, str, Dict]:
        apikey, device_id, payload = super().decode_message(msg=msg, decoder=decoder)
        payload = payload.split("@")
        if not device_id == payload[0]:
            self.logger.warning("Received invalid command")

        payload = payload[1].split("|")
        payload = {
            payload[i]: self.__eval_value(payload[i + 1])
            for i in range(0, len(payload), 2)
        }

        return apikey, device_id, payload

    def encode_msg(
        self, device_id: str, payload: Any, msg_type: IoTAMQTTMessageType
    ) -> str:
        if msg_type == IoTAMQTTMessageType.SINGLE:
            return payload
        elif msg_type == IoTAMQTTMessageType.MULTI:
            payload = super()._parse_timestamp(payload=payload)
            timestamp = str(payload.pop("timeInstant", ""))
            data = "|".join([f"{key}|{value}" for key, value in payload.items()])
            data = "|".join([timestamp, data]).strip("|")
            return data
        elif msg_type == IoTAMQTTMessageType.CMDEXE:
            for key, value in payload.items():
                if isinstance(value, bool):
                    value = str(value).lower()
                elif isinstance(value, (float, int)):
                    value = str(value)
                elif isinstance(value, str):
                    pass
                else:
                    raise ValueError("Cannot parse command acknowledgement!")
                return f"{device_id}@{key}|{value}"
        super()._raise_encoding_error(payload=payload, msg_type=msg_type)



"""
Variable types and classes used for better validation
"""

from pydantic import UrlConstraints
from typing_extensions import Annotated
from pydantic_core import Url

AnyMqttUrl = Annotated[Url, UrlConstraints(allowed_schemes=["mqtt"])]



"""Manages the local state of the semantic instances"""

import copy
import json
import logging
import uuid
from math import inf

import requests

from typing import Optional, Dict, Type, List, Any, Union, Set
from pydantic import BaseModel, Field
from rapidfuzz import process

from filip.models.base import NgsiVersion
from filip.models.ngsi_v2.iot import DeviceSettings
from filip.semantics.vocabulary import Individual
from filip.models.ngsi_v2.context import ContextEntity
from filip.clients.ngsi_v2 import ContextBrokerClient, IoTAClient
from filip.models import FiwareHeader
from filip.semantics.semantics_models import (
    InstanceIdentifier,
    SemanticClass,
    InstanceHeader,
    Datatype,
    DataField,
    RelationField,
    SemanticIndividual,
    SemanticDeviceClass,
    CommandField,
    Command,
    DeviceAttributeField,
    DeviceAttribute,
)
from filip.utils.simple_ql import QueryString


logger = logging.getLogger("semantics")


class InstanceRegistry(BaseModel):
    """
    Holds all the references to the local SemanticClass instances.
    The instance registry is a global object, that is directly inject in the
    SemanticClass constructor over the SemanticsManager
    """

    _registry: Dict[InstanceIdentifier, "SemanticClass"] = {}
    """ Dict of the references to the local SemanticClass instances. 
        Instances are saved with their identifier as key """

    _deleted_identifiers: List[InstanceIdentifier] = []
    """List of all identifiers that were deleted"""

    def delete(self, instance: "SemanticClass"):
        """Delete an instance from the registry

        Args:
            instance(SemanticClass): Instance to remove

        Raises:
           KeyError, if identifier unknown

        Returns:
            None
        """
        identifier = instance.get_identifier()
        if not self.contains(identifier):
            raise KeyError(f"Identifier {identifier} unknown, " f"can not delete")

        # If instance was loaded from Fiware it has an old_state.
        # if that is the case, we need to note that we have deleted the instance
        # to delete it on save, and do not load it again from Fiware

        if instance.old_state.state is not None:
            self._deleted_identifiers.append(identifier)

        del self._registry[identifier]

    def instance_was_deleted(self, identifier: InstanceIdentifier) -> bool:
        """
        Check if an instance was deleted

        Args:
            identifier (InstanceIdentifier): Identifier of instance to check

        Returns:
            bool
        """
        return identifier in self._deleted_identifiers

    def register(self, instance: "SemanticClass"):
        """
        Register a new instance of a SemanticClass in the registry

        Args:
            instance(SemanticClass):  Instance to be registered
        Raises:
            AttributeError: if Instance is already registered
        """
        identifier = instance.get_identifier()

        if identifier in self._registry:
            raise AttributeError("Instance already exists")
        else:
            self._registry[identifier] = instance

    def get(self, identifier: InstanceIdentifier) -> "SemanticClass":
        """Retrieve an registered instance with its identifier

        Args:
            identifier(InstanceIdentifier): identifier belonging to instance
        Returns:
            SemanticClass
        """
        return self._registry[identifier]

    def contains(self, identifier: InstanceIdentifier) -> bool:
        """Test if an identifier is registered

        Args:
            identifier(InstanceIdentifier): identifier belonging to instance
        Returns:
            bool, True if registered
        """
        return identifier in self._registry

    def get_all(self) -> List["SemanticClass"]:
        """Get all registered instances

        Returns:
            List[SemanticClass]
        """
        return list(self._registry.values())

    def get_all_deleted_identifiers(self) -> List["InstanceIdentifier"]:
        """
        Get all identifiers that were deleted by the user

        Returns:
            List[InstanceIdentifier]
        """
        return self._deleted_identifiers

    def save(self) -> str:
        """
        Save the state of the registry out of a json string.

        Returns:
             str, json string of registry state
        """
        res = {"instances": [], "deleted_identifiers": []}

        for identifier, instance in self._registry.items():
            old_state = None
            if instance.old_state.state is not None:
                old_state = instance.old_state.state.model_dump_json()
            instance_dict = {
                "entity": instance.build_context_entity().model_dump_json(),
                "header": instance.header.model_dump_json(),
                "old_state": old_state,
            }
            res["instances"].append(instance_dict)

        for identifier in self._deleted_identifiers:
            res["deleted_identifiers"].append(identifier.model_dump_json())

        return json.dumps(res, indent=4)

    def clear(self):
        """Clear the local state"""
        self._registry.clear()
        self._deleted_identifiers.clear()

    def load(self, json_string: str, semantic_manager: "SemanticsManager"):
        """
        Load the state of the registry out of a json string. The current
        state will be discarded

        Args:
            json_string (str): State expressed as json string
            semantic_manager (SemanticsManager): manager to which registry
                belongs
        Returns:
             None
        """
        self.clear()

        save = json.loads(json_string)
        for instance_dict in save["instances"]:
            entity_json = instance_dict["entity"]
            header = InstanceHeader.model_validate(instance_dict["header"])

            context_entity = ContextEntity.model_validate(entity_json)

            instance = semantic_manager._context_entity_to_semantic_class(
                context_entity, header
            )

            if instance_dict["old_state"] is not None:
                instance.old_state.state = ContextEntity.model_validate(
                    instance_dict["old_state"]
                )

            self._registry[instance.get_identifier()] = instance

        for identifier in save["deleted_identifiers"]:
            self._deleted_identifiers.append(
                InstanceIdentifier.model_validate(identifier)
            )

    def __hash__(self):
        values = (hash(value) for value in self._registry.values())

        return hash((frozenset(values), frozenset(self._deleted_identifiers)))


class SemanticsManager(BaseModel):
    """
    The Semantic Manager is a static object that is delivered with
    each vocabulary model export.

    It provides the interface to interact with the local state and Fiware
    """

    instance_registry: InstanceRegistry = Field(
        description="Registry managing the local state"
    )
    class_catalogue: Dict[str, Type[SemanticClass]] = Field(
        default={}, description="Register of class names to classes"
    )
    datatype_catalogue: Dict[str, Dict[str, str]] = Field(
        default={},
        description="Register of datatype names to Dict representation of " "datatypes",
    )
    individual_catalogue: Dict[str, type] = Field(
        default={}, description="Register of individual names to their classes"
    )

    default_header: InstanceHeader = Field(
        default=InstanceHeader(),
        description="Default header that each new instance receives if it "
        "does not specify an own header",
    )

    @staticmethod
    def get_client(instance_header: InstanceHeader) -> ContextBrokerClient:
        """Get the correct ContextBrokerClient to be used with the given header

        Args:
            instance_header (InstanceHeader): Header to be used with client
        Returns:
            ContextBrokerClient
        """
        if instance_header.ngsi_version == NgsiVersion.v2:
            return ContextBrokerClient(
                url=instance_header.cb_url,
                fiware_header=instance_header.get_fiware_header(),
            )
        else:
            # todo LD
            raise Exception("FiwareVersion not yet supported")

    @staticmethod
    def get_iota_client(instance_header: InstanceHeader) -> IoTAClient:
        """Get the correct IotaClient to be used with the given header

        Args:
            instance_header (InstanceHeader): Header to be used with client
        Returns:
            IoTAClient
        """
        if instance_header.ngsi_version == NgsiVersion.v2:
            return IoTAClient(
                url=instance_header.iota_url,
                fiware_header=instance_header.get_fiware_header(),
            )
        else:
            # todo LD
            raise Exception("FiwareVersion not yet supported")

    def _context_entity_to_semantic_class(
        self, entity: ContextEntity, header: InstanceHeader
    ) -> SemanticClass:
        """Converts a ContextEntity to a SemanticClass

        Args:
            entity (ContextEntity): entity to convert
            header (InstanceHeader): Header of the new instance

        Returns:
            SemanticClass or SemanticDeviceClass
        """

        class_name = entity.type

        class_: Type = self.get_class_by_name(class_name)

        if not self.is_class_name_an_device_class(class_name):

            loaded_class: SemanticClass = class_(
                id=entity.id, header=header, enforce_new=True
            )
        else:
            loaded_class: SemanticDeviceClass = class_(
                id=entity.id, header=header, enforce_new=True
            )

        loaded_class.old_state.state = entity

        # load values of class from the context_entity into the instance
        for field in loaded_class.get_fields():
            field.clear()  # remove default values, from hasValue relations
            field_name = field.name
            entity_attribute = entity.get_attribute(field_name)
            if entity_attribute is None:
                raise Exception(
                    f"The corresponding entity for ({entity.id},{entity.type}) "
                    f"in Fiware misses a field that "
                    f"is required by the class_model: {field_name}. The "
                    f"fiware state and the used vocabulary models are not "
                    f"compatible"
                )

            entity_field_value = entity.get_attribute(field_name).value

            if isinstance(entity_field_value, List):
                values = entity_field_value
            else:
                values = [entity_field_value]

            for value in values:
                converted_value = self._convert_value_fitting_for_field(field, value)
                if isinstance(field, RelationField):
                    # we need to bypass the main setter, as it expects an
                    # instance and we do not want to load the instance if it
                    # is not used
                    field._set.add(converted_value)
                else:
                    field.add(converted_value)

        # load references into instance
        references_attribute = entity.get_attribute("referencedBy")
        references = references_attribute.value

        for identifier_str, prop_list in references.items():
            for prop in prop_list:
                loaded_class.add_reference(
                    InstanceIdentifier.model_validate_json(
                        identifier_str.replace("---", ".")
                    ),
                    prop,
                )

        # load metadata
        metadata_dict = entity.get_attribute("metadata").value
        loaded_class.metadata.name = metadata_dict["name"]
        loaded_class.metadata.comment = metadata_dict["comment"]

        # load device_settings into instance, if instance is a device
        if isinstance(loaded_class, SemanticDeviceClass):
            settings_attribute = entity.get_attribute("deviceSettings")
            device_settings = DeviceSettings.model_validate(settings_attribute.value)

            for key, value in device_settings.model_dump().items():
                loaded_class.device_settings.__setattr__(key, value)

        return loaded_class

    @staticmethod
    def _convert_value_fitting_for_field(field, value):
        """
        Converts a given value into the correct format for the given field

        Args:
            field: SemanticField
            value: Value to convert

        Returns:
            converted value
        """
        if isinstance(field, DataField):
            return value
        elif isinstance(field, RelationField):
            # convert json to Identifier, inject identifier in Relation,
            # the class will be hotloaded if the value in the is
            # relationship is accessed

            if not isinstance(value, dict):  # is an individual
                return value
            else:  # is an instance_identifier
                # we need to replace back --- with . that we switched,
                # as a . is not allowed in the dic in Fiware
                return InstanceIdentifier.model_validate_json(
                    str(value).replace("---", ".").replace("'", '"')
                )

        elif isinstance(field, CommandField):
            if isinstance(value, Command):
                return value
            # if loading local state, the wrong string delimters are used,
            # and the string is not automatically converted to a dict
            if not isinstance(value, dict):
                value = json.loads(value.replace("'", '"'))

            return Command(name=value["name"])
        elif isinstance(field, DeviceAttributeField):

            # if loading local state, the wrong string delimters are used,
            # and the string is not automatically converted to a dict

            if isinstance(value, DeviceAttribute):
                return value
            if not isinstance(value, dict):
                value = json.loads(value.replace("'", '"'))

            return DeviceAttribute(
                name=value["name"], attribute_type=value["attribute_type"]
            )

    def get_class_by_name(self, class_name: str) -> Type[SemanticClass]:
        """
        Get the class object by its type in string form

        Args:
            class_name (str)

        Raises:
            KeyError: if class_name not registered as a SemanticClass

        Returns:
            Type
        """
        return self.class_catalogue[class_name]

    def is_class_name_an_device_class(self, class_name: str) -> bool:
        """
        Test if the name/type of a class belongs to a SemanticDeviceClass

        Args:
            class_name (str): class name to check

        Returns:
            bool, True if belongs to a SemanticDeviceClass
        """
        class_type = self.get_class_by_name(class_name)
        return isinstance(class_type, SemanticDeviceClass)

    def is_local_state_valid(self, validate_rules: bool = True) -> (bool, str):
        """
        Check if the local state is valid and can be saved.

        Args:
            validate_rules (bool): If true Rulefields are validated

        Returns:
            (bool, str): (Is valid?, Message)
        """

        if validate_rules:
            for instance in self.instance_registry.get_all():
                if isinstance(instance, Individual):
                    continue
                if not instance.are_rule_fields_valid():
                    return (
                        False,
                        f"SemanticEntity {instance.id} of type"
                        f"{instance.get_type()} has unfulfilled fields "
                        f"{[f.name for f in instance.get_invalid_rule_fields()]}.",
                    )

        for instance in self.instance_registry.get_all():
            if isinstance(instance, SemanticDeviceClass):
                if instance.device_settings.transport is None:
                    return (
                        False,
                        f"Device {instance.id} of type {instance.get_type()} "
                        f"needs to be given an transport setting.",
                    )
        return True, "State is valid"

    def save_state(self, assert_validity: bool = True):
        """
        Save the local state completely to Fiware.

        Args:
            assert_validity (bool): It true an error is raised if the
            RuleFields of one instance are invalid

        Raises:
            AssertionError: If a device endpoint or transport is not defined

        Returns:
            None
        """
        (valid, msg) = self.is_local_state_valid(validate_rules=assert_validity)

        if not valid:
            raise AssertionError(f"{msg}. Local state was not saved")

        # delete all instance that were loaded from Fiware and then deleted
        # wrap in try, as the entity could have been deleted by a third party
        for identifier in self.instance_registry.get_all_deleted_identifiers():

            # we need to handle devices and normal classes with different
            # clients

            client = self.get_client(instance_header=identifier.header)
            iota_client = self.get_iota_client(instance_header=identifier.header)
            try:
                client.delete_entity(
                    entity_id=identifier.id,
                    entity_type=identifier.type,
                    delete_devices=True,
                    iota_client=iota_client,
                )
            except requests.RequestException:
                raise

            client.close()

        # merge with live state
        for instance in self.instance_registry.get_all():
            self.merge_local_and_live_instance_state(instance)

        # save, patch all local instances
        for instance in self.instance_registry.get_all():
            cb_client = self.get_client(instance_header=instance.header)
            if not isinstance(instance, SemanticDeviceClass):
                # it is important that we patch the values else the
                # references field would reach an invalid state if we worked
                # in parallel on an instance
                cb_client.patch_entity(
                    instance.build_context_entity(), instance.old_state.state
                )
            else:
                iota_client = self.get_iota_client(instance_header=instance.header)
                iota_client.patch_device(
                    device=instance.build_context_device(),
                    patch_entity=True,
                    cb_client=cb_client,
                )
                iota_client.close()
            cb_client.close()
        # update old_state
        for instance in self.instance_registry.get_all():
            instance.old_state.state = instance.build_context_entity()

    def load_instance(self, identifier: InstanceIdentifier) -> SemanticClass:
        """
        Get the instance with the given identifier. It is either loaded from
        local state or retrieved from fiware

        Args:
            identifier (InstanceIdentifier): Identifier to load

        Returns:
            SemanticClass
        """

        if self.instance_registry.contains(identifier=identifier):
            return self.instance_registry.get(identifier=identifier)
        else:
            client = self.get_client(identifier.header)

            entity = client.get_entity(
                entity_id=identifier.id, entity_type=identifier.type
            )
            client.close()

            logger.info(
                f"Instance ({identifier.id}, {identifier.type}) "
                f"loaded from Fiware({identifier.header.cb_url}"
                f", {identifier.header.service}"
                f"{identifier.header.service_path})"
            )
            return self._context_entity_to_semantic_class(
                entity=entity, header=identifier.header
            )

    def does_instance_exists(self, identifier: InstanceIdentifier) -> bool:
        """
        Check if an instance with the given identifier already exists in
        local state or in Fiware

        Args:
            identifier (InstanceIdentifier): Identifier to check

        Returns:
            bool, true if exists
        """

        if self.instance_registry.contains(identifier=identifier):
            return True
        elif self.was_instance_deleted(identifier):
            return False
        else:
            client = self.get_client(identifier.header)
            return client.does_entity_exist(
                entity_id=identifier.id, entity_type=identifier.type
            )

    def was_instance_deleted(self, identifier: InstanceIdentifier) -> bool:
        """
        Check if the instance with the given identifier was deleted.

        Args:
            identifier (InstanceIdentifier): Identifier to check

        Returns:
            bool, true if deleted
        """
        return self.instance_registry.instance_was_deleted(identifier)

    def get_instance(self, identifier: InstanceIdentifier) -> SemanticClass:
        """
        Get the instance with the given identifier. It is either loaded from
        local state or retrieved from fiware

        Args:
            identifier (InstanceIdentifier): Identifier to load

        Returns:
            SemanticClass
        """
        return self.load_instance(identifier)

    def get_all_local_instances(self) -> List[SemanticClass]:
        """
        Retrieve all SemanticClass instances in the local state

        Returns:
            List[SemanticClass]
        """
        return self.instance_registry.get_all()

    def get_all_local_instances_of_class(
        self,
        class_: Optional[type] = None,
        class_name: Optional[str] = None,
        get_subclasses: bool = True,
    ) -> List[SemanticClass]:
        """
        Retrieve all instances of a SemanitcClass from Local Storage

        Args:
            class_ (type):
                Type of classes to retrieve
            class_name (str):
                Name of type of classes to retrieve as string
            get_subclasses (bool):
                If true also all instances of subclasses
                of given class are returned

        Raises:
            AssertionError: If class_ and class_name are both None or non None

        Returns:
            List[SemanticClass]
        """

        assert class_ is None or class_name is None, "Only one parameter is allowed"
        assert class_ is not None or class_name is not None, "One parameter is required"

        if class_ is not None:
            class_name = class_.__name__
        else:
            class_ = self.get_class_by_name(class_name)

        res = []
        for instance in self.instance_registry.get_all():
            if not get_subclasses:
                if instance.get_type() == class_name:
                    res.append(instance)
            else:
                if isinstance(instance, class_):
                    res.append(instance)
        return res

    def load_instances_from_fiware(
        self,
        fiware_header: FiwareHeader,
        fiware_version: NgsiVersion,
        cb_url: str,
        iota_url: str,
        entity_ids: Optional[List[str]] = None,
        entity_types: Optional[List[str]] = None,
        id_pattern: str = None,
        type_pattern: str = None,
        q: Union[str, QueryString] = None,
        limit: int = inf,
    ) -> List[SemanticClass]:
        """
        Loads the instances of given types or ids from Fiware into the local
        state and returns the loaded instances

        Args:
            fiware_header (FiwareHeader): Fiware location to load
            fiware_version (NgsiVersion): Used fiware version
            cb_url (str): URL of the ContextBroker
            iota_url (str): URL of the IotaBroker
            entity_ids (Optional[str]): List of the entities ids that
                should be loaded
            entity_types (Optional[str]): List of the entities types that
                should be loaded
            id_pattern: A correctly formatted regular expression. Retrieve
                entities whose ID matches the regular expression. Incompatible
                with id, e.g. ngsi-ld.* or sensor.*
            type_pattern: A correctly formatted regular expression. Retrieve
                entities whose type matches the regular expression.
                Incompatible with type, e.g. room.*
            q (SimpleQuery): A query expression, composed of a list of
                statements separated by ;, i.e.,
                q=statement1;statement2;statement3. See Simple Query
                Language specification. Example: temperature>40.
            limit: Limits the number of entities to be retrieved Example: 20

        Raises:
           ValueError: if both entity_types and entity_ids are given
           ValueError: if Retrival of Context-entities fails
        Returns:
             List[SemanticClass]
        """

        header: InstanceHeader = InstanceHeader(
            service=fiware_header.service,
            service_path=fiware_header.service_path,
            cb_url=cb_url,
            iota_url=iota_url,
            ngsi_version=fiware_version,
        )

        client = self.get_client(header)

        entities = client.get_entity_list(
            entity_ids=entity_ids,
            entity_types=entity_types,
            id_pattern=id_pattern,
            type_pattern=type_pattern,
            q=q,
            limit=limit,
        )
        client.close()

        return [self._context_entity_to_semantic_class(e, header) for e in entities]

    def get_entity_from_fiware(
        self, instance_identifier: InstanceIdentifier
    ) -> ContextEntity:
        """
        Retrieve the current entry of an instance in Fiware

        Args:
            instance_identifier (InstanceIdentifier): Identifier to load

        Raises:
            Exception, if Entity is not present

        Returns:
              ContextEntity
        """
        client = self.get_client(instance_identifier.header)

        return client.get_entity(
            entity_id=instance_identifier.id, entity_type=instance_identifier.type
        )

    def load_instances(
        self, identifiers: List[InstanceIdentifier]
    ) -> List[SemanticClass]:
        """
        Load all instances, if no local state of it exists it will get taken
        from Fiware and registered locally

        Args:
            identifiers List[InstanceIdentifier]: Identifiers of instances
                that should be loaded
        Raises:
            Exception, if one Entity is not present

        Returns:
           List[SemanticClass]
        """

        return [self.load_instance(iden) for iden in identifiers]

    def set_default_header(self, header: InstanceHeader):
        """
        Set the default header, which all new instance that does not specify a
        header in the constructor receives

        Args:
            header (InstanceHeader): new default header

        Returns:
            None
        """
        self.default_header = copy.deepcopy(header)

    def get_default_header(self) -> InstanceHeader:
        """
        Instance header is read-only, therefore giving back a copy is
        theoretically not needed, but it is cleaner that all instance has an
        own header object that is not shared
        """
        return copy.deepcopy(self.default_header)

    def get_datatype(self, datatype_name: str) -> Datatype:
        """
        Get a Datatype object with the name as key as specified in the model

        Args:
            datatype_name (str): key label of the datatype

        Returns:
            Datatype
        """
        return Datatype.model_validate(self.datatype_catalogue[datatype_name])

    def get_individual(self, individual_name: str) -> SemanticIndividual:
        """
        Get an individual by its name

        Args:
            individual_name (str)
        Raises:
            KeyError, if name not registered
        Returns:
            SemanticIndividual
        """
        return self.individual_catalogue[individual_name]()

    def save_local_state_as_json(self) -> str:
        """
        Save the local state with all made changes as json string

        Returns:
            Json String, containing all information about the local state
        """
        return self.instance_registry.save()

    def load_local_state_from_json(self, json: str):
        """
        Loads the local state from a json string. The current local state gets
        discarded

        Raises:
            Error, if not a correct json string

        """
        self.instance_registry.load(json, self)

    def visualize_local_state(self, display_individuals_rule: str = "ALL"):
        """
        Visualise all instances in the local state in a network graph that
        shows which instances reference each other over which fields

        On execution of the methode a temporary image file is created and
        automatically displayed in the standard image viewing software of the
        system

        Args:
            display_individuals_rule (rule):
                If:
                "USED": Show only Individuals
                "ALL": Display all known Individuals
                "NONE": Display no Individuals
                that are connected to at least one instance
                else: Show all individuals

        Raises:
            ValueError: if display_individuals_rule is invalid
        """

        if (
            not display_individuals_rule == "ALL"
            and not display_individuals_rule == "NONE"
            and not display_individuals_rule == "USED"
        ):

            raise ValueError(f"Invalid parameter {display_individuals_rule}")

        import igraph

        g = igraph.Graph(directed=True)

        for instance in self.get_all_local_instances():
            g.add_vertex(
                name=instance.id,
                label=f"\n\n\n {instance.get_type()} \n {instance.id}",
                color="green",
            )

        used_individuals_names: Set[str] = set()
        for instance in self.get_all_local_instances():
            for field in instance.get_relation_fields():
                for linked in field.get_all():
                    if isinstance(linked, SemanticClass):
                        g.add_edge(instance.id, linked.id, name=field.name)
                        # g.es[-1]["name"] = field.name

                    elif isinstance(linked, SemanticIndividual):
                        if not display_individuals_rule == "NONE":
                            g.add_edge(instance.id, linked.get_name())
                            used_individuals_names.add(linked.get_name())

        if display_individuals_rule == "ALL":
            used_individuals_names.update(self.individual_catalogue.keys())
        for individual in [
            self.get_individual(name) for name in used_individuals_names
        ]:
            g.add_vertex(
                label=f"\n\n\n{individual.get_name()}",
                name=individual.get_name(),
                color="blue",
            )

        layout = g.layout("fr")
        visual_style = {
            "vertex_size": 20,
            "vertex_color": g.vs["color"],
            "vertex_label": g.vs["label"],
            "edge_label": g.es["name"],
            "layout": layout,
            "bbox": (len(g.vs) * 50, len(g.vs) * 50),
        }

        igraph.plot(g, **visual_style)

    def generate_cytoscape_for_local_state(
        self, display_only_used_individuals: bool = True
    ):
        """
        Generate a graph definition that can be loaded into a cytoscape
        visualisation tool, that describes the complete current local state.

        For the graph layout COLA is recommended with an edge length of 150

        Args:
            display_only_used_individuals (bool):
                If true(default): Show only Individuals that are connected to
                at least one instance
                else: Show all individuals

        Returns:
            Tupel of elements and stylesheet:
                elements is a dict:
                {"nodes": NODE_DEFINITIONS, "edges": EDGE_DEFINITIONS}
                stylesheet is a list containing all the graph styles
        """

        # graph design
        stylesheet = [
            {"selector": "node", "style": {"label": "data(label)", "z-index": 9999}},
            {
                "selector": "edge",
                "style": {
                    "curve-style": "bezier",
                    "target-arrow-color": "black",
                    "target-arrow-shape": "triangle",
                    "line-color": "black",
                    "opacity": 0.45,
                    "z-index": 5000,
                },
            },
            {
                "selector": ".center",
                "style": {"shape": "rectangle", "background-color": "black"},
            },
            {
                "selector": ".individual",
                "style": {"shape": "circle", "background-color": "orange"},
            },
            {
                "selector": ".instance",
                "style": {"shape": "circle", "background-color": "green"},
            },
            {
                "selector": ".collection",
                "style": {"shape": "triangle", "background-color": "gray"},
            },
        ]

        nodes = []
        edges = []

        used_individual_names = set()
        if not display_only_used_individuals:
            used_individual_names.update(self.individual_catalogue.keys())

        def get_node_id(item: Union[SemanticClass, SemanticIndividual]) -> str:
            """
            Get the id to be used in the graph for an item

            Args:
                item (Union[SemanticClass, SemanticIndividual]): Item to get
                                                                    ID for

            Returns:
                str - ID
            """
            if isinstance(item, SemanticIndividual):
                return item.get_name()
            else:
                return item.get_identifier().model_dump_json()

        for instance in self.get_all_local_instances():
            label = f"({instance.get_type()}){instance.metadata.name}"
            nodes.append(
                {
                    "data": {
                        "id": get_node_id(instance),
                        "label": label,
                        "parent_id": "",
                        "classes": "instance item",
                    },
                    "classes": "instance item",
                }
            )

        for instance in self.get_all_local_instances():

            for rel_field in instance.get_relation_fields():

                values = rel_field.get_all()
                for v in values:
                    if isinstance(v, SemanticIndividual):
                        used_individual_names.add(v.get_name())

                if len(values) == 0:
                    pass
                elif len(values) == 1:
                    edge_id = uuid.uuid4().hex
                    edges.append(
                        {
                            "data": {
                                "id": edge_id,
                                "source": get_node_id(instance),
                                "target": get_node_id(values[0]),
                            }
                        }
                    )
                    edge_name = rel_field.name
                    stylesheet.append(
                        {"selector": "#" + edge_id, "style": {"label": edge_name}}
                    )
                else:
                    edge_id = uuid.uuid4().hex
                    node_id = uuid.uuid4().hex
                    nodes.append(
                        {
                            "data": {
                                "id": node_id,
                                "label": "",
                                "parent_id": "",
                                "classes": "collection",
                            },
                            "classes": "collection",
                        }
                    )

                    edges.append(
                        {
                            "data": {
                                "id": edge_id,
                                "source": get_node_id(instance),
                                "target": node_id,
                            }
                        }
                    )
                    edge_name = rel_field.name
                    stylesheet.append(
                        {"selector": "#" + edge_id, "style": {"label": edge_name}}
                    )

                    for value in values:
                        edge_id = uuid.uuid4().hex
                        edges.append(
                            {
                                "data": {
                                    "id": edge_id,
                                    "source": node_id,
                                    "target": get_node_id(value),
                                }
                            }
                        )

        for individual_name in used_individual_names:
            nodes.append(
                {
                    "data": {
                        "id": individual_name,
                        "label": individual_name,
                        "parent_id": "",
                        "classes": "individual item",
                    },
                    "classes": "individual item",
                }
            )

        elements = {"nodes": nodes, "edges": edges}

        return elements, stylesheet

    def merge_local_and_live_instance_state(self, instance: SemanticClass) -> None:
        """
        The live state of the instance is fetched from Fiware (if it exists)
        and the two states are merged:

        For each Field:
        - each added value (compared to old_state) is added to
        the live state
        - each deleted value (compared to old_state) is removed from
        the live state

        For each Device Settings (if instance is device):
        - If the device setting changed (compared to old_state) the live
        setting is overwritten

        For each Reference:
        - each added value (compared to old_state) is added to
        the live state
        - each deleted value (compared to old_state) is removed from
        the live state

        The new state is directly saved in the instance

        Args:
              instance (SemanticClass): instanced to be treated
        """

        def converted_attribute_values(field, attribute) -> Set:
            return {
                self._convert_value_fitting_for_field(field, value)
                for value in attribute.value
            }

        def _get_added_and_removed_values(
            old_values: Union[List, Set, Any], current_values: Union[List, Set, Any]
        ) -> (Set, Set):

            old_set = set(old_values)
            current_set = set(current_values)
            added_values = set()
            removed_values = set()

            # remove deleted values from live state, it can be that the value
            # was also deleted in the live state
            for value in old_set:
                if value not in current_set:
                    removed_values.add(value)

            # add added values
            for value in current_set:
                if value not in old_set:
                    added_values.add(value)

            return added_values, removed_values

        # instance is new. Save it as is
        client = self.get_client(instance.header)
        if not client.does_entity_exist(
            entity_id=instance.id, entity_type=instance.get_type()
        ):
            return

        client = self.get_client(instance.header)
        live_entity = client.get_entity(
            entity_id=instance.id, entity_type=instance.get_type()
        )
        client.close()

        current_entity = instance.build_context_entity()
        old_entity = instance.old_state.state

        # ------merge fields-----------------------------------------------
        # instance exists already, add all locally added and delete all
        # locally deleted values to the/from the live_state

        for field in instance.get_fields():
            # live_values = set(live_entity.get_attribute(field.name).value)
            live_values = converted_attribute_values(
                field, live_entity.get_attribute(field.name)
            )
            old_values = converted_attribute_values(
                field, old_entity.get_attribute(field.name)
            )
            current_values = converted_attribute_values(
                field, current_entity.get_attribute(field.name)
            )

            (added_values, deleted_values) = _get_added_and_removed_values(
                old_values,
                current_values,
                # old_entity.get_attribute(field.name).value,
                # current_entity.get_attribute(field.name).value
            )

            for value in added_values:
                live_values.add(value)
            for value in deleted_values:
                if value in live_values:
                    live_values.remove(value)

            new_values = list(live_values)
            # update local stated with merged result
            field._set.clear()  # very important to not use field.clear,
            # as that methode would also delete references
            for value in new_values:
                converted_value = self._convert_value_fitting_for_field(field, value)
                field._set.add(converted_value)

        # ------merge references-----------------------------------------------
        merged_references: Dict = live_entity.get_attribute("referencedBy").value
        current_references: Dict = current_entity.get_attribute("referencedBy").value
        old_references: Dict = old_entity.get_attribute("referencedBy").value

        keys = set(current_references.keys())
        keys.update(old_references.keys())

        for key in keys:
            current_values = []
            old_values = []
            if key in current_references:
                current_values = current_references[key]
            if key in old_references:
                old_values = old_references[key]

            (added_values, deleted_values) = _get_added_and_removed_values(
                current_values=current_values, old_values=old_values
            )

            # ensure the merged state has each key
            if key not in merged_references.keys():
                merged_references[key] = []

            # add, added values that did not exist before
            for value in added_values:
                if value not in merged_references[key]:
                    merged_references[key].append(value)

            # delete deleted values if they were not already deleted
            for value in deleted_values:
                if value in merged_references[key]:
                    merged_references[key].remove(value)

            # delete all keys that point to empty lists
            keys_to_delete = []
            for key, value in merged_references.items():
                if len(value) == 0:
                    keys_to_delete.append(key)
            for key in keys_to_delete:
                del merged_references[key]

        # save merged references
        instance.references.clear()
        for key, value in merged_references.items():
            # replace back the protected . (. not allowed in keys in fiware)
            instance.references[
                InstanceIdentifier.model_validate_json(key.replace("---", "."))
            ] = value

        # ------merge device settings----------------------------------------
        if isinstance(instance, SemanticDeviceClass):
            old_settings = old_entity.get_attribute("deviceSettings").value
            current_settings = current_entity.get_attribute("deviceSettings").value
            new_settings = live_entity.get_attribute("deviceSettings").value

            # keys are always the same
            # override live state with local changes
            for key in old_settings:
                if old_settings[key] is not current_settings[key]:
                    new_settings[key] = current_settings[key]
                instance.device_settings.__setattr__(key, new_settings[key])

    def find_fitting_model(self, search_term: str, limit: int = 5) -> List[str]:
        """
        Find a fitting model by entering a search_term (e.g.: Sensor).
        The methode returns a selection from up-to [limit] possibly fitting
        model names. If a model name was selected from the proposition the
        model can be retrieved with the methode:
        "get_class_by_name(selectedName)"

        Args:
            search_term (str): search term to find a model by name
            limit (int): Max Number of suggested results (default: 5)

        Returns:
            List[str], containing 0 to [limit] ordered propositions (best first)
        """
        class_names = list(self.class_catalogue.keys())
        suggestions = [
            item[0]
            for item in process.extract(
                query=search_term.casefold(),
                choices=class_names,
                score_cutoff=50,
                limit=limit,
            )
        ]

        return suggestions
